# Author: Anthony Stahler
# Last updated: Jan 5, 2023
# License: Free for personal use, for commercial use please reach out to the author via the BestEdit contact page: http://bestedit.net/contact/
# The most recent version of this script can be found at http://bestedit.net/premiere-pro-xml-fix/
#
# This code is the result of a lot of testing and rewriting, so it's a bit of a mess and I don't want to clean it up and inadvertently re-add an edge-case bug I've fixed during testing. 
# So if it sometimes processes data twice for certain tags, sorry, there may be an edge case reason or it could just be a legacy code inefficiency from previous iterations that was later corrected in a different spot. 
# But since it doesn't harm anything no reason to risk breaking it for cleaner code.
#
# Here's a rough description of the changes between versions:
#
# Version 2: fixed media name cleaning so it doesn't break "Relink others automatically" in Premiere
# Version 3: fixed script bug that could make first clip on timeline disappear in some cases, and a Premiere bug where START transitions are sometimes marked as CENTER
# Version 4: use ticks instead of framerate for slightly better precision, small tweaks to fix certain clips that can be shifted off a frame or two by the script, depending on transition used
# Version 5: fixed a bug in the script's transition length calculation, added code to reject full project xml files to avoid bugs
# Version 6: Added the ability to pre-process the full data with regex. Fixed a bug where END-BLACK (one sided) transitions disappear if rounding results in a length different than the cutpointticks. Also added support for speed/duration adjustments by converting it to time remapping properties on the clips.
# Version 7: Added a setting (default: on) for simulated Exponential Fades replacing the Constant Gain transition type, since Constant Power and Exponential Fade are the most powerful audio transitions and for some reason Premiere incorrectly converts EF to CG on export
# Version 8: Added source video tracking so that the video ID number is consistent across XML projects for any given video
# Version 9: Small code corrections
# Version 10: Remove a cutpointticks fudge that was causing an audio/video sync error, fixed a mismatch between OUT and pproTicksOut values
# Version 11: Toyed with readding END alignment to maintain consistency in projects when framerate is changed to fix simulated END transitions centering themselves incorrectly when changing framerate. However this often results in the left or right side of transitions disconnecting from the transition, effectively turning it into a START-BLACK or END-BLACK. Instead, I'm adding support for 60fps, 25fps and 23.976fps, and converting to the closest safe framerate.  Should be easy to add other framerates to the $fpsarr/$timebasearr/$ntscarr arrays if you need them too.
# Version 12: Added four more framerates to the $fpsarr/$timebasearr/$ntscarr arrays and clearer notes on how to convert your own.
# Version 13: Added the option to re-center transitions for better Premiere compatibility, non-center can sometimes change position when copied
#
# When exporting your XML, click to highlight the desired sequence in the project tab before doing File > Export > Final Cut Pro XML, this will export just the sequence and not the entire project which may have multiple sequences. 
# Exporting an entire project may work but is not tested.
#
#
# Please note, this will fix the ability to import xml files but may make slight alterations to your edit to work around Adobe Premiere Pro export and import bugs.
# In most situations these slight alterations shouldn't be noticable, but if you're editing with very tight tolerances where every single frame matters, it could have an impact.
# If that's the case you're probably better off just sharing the prproj, although it does contain some personally identifable info about your configuration, file paths on your PC, etc.
#
# Even though you can import dolby 7.1 audio into premiere it doesn't respect the export properly in the xml file, it turns it into multiple stereo and mono tracks that are impossible to distinguish from normal stereo/mono in xml.  
# If you use the script please use ffmpeg to convert your 7.1 file to 5.1.  If you edit and export it as 7.1 there is no way to fix the 2021 and prior export files.
#
# List of issues fixed:
# The total video length shown by Premiere Pro 2021 is incorrect for many video framerates, the reimported project shows the correct timeline length
#
# PP 2019 and 2021 handle framerates differently, 2021 uses 6 digits after the decimal while 2019 only uses 3 digits. This causes bugs in the 2021 timeline with transition lengths rounding improperly depending on the video file framerate.
#
# Both versions of Premiere Pro also have a bug where any transition which is shifted all the way to the left so it only overlaps the left audio clip and not the right one will not import correctly
# In the xml this is an END transition, and to fix the bug we shift the cut in the audio one frame to the left, so the transition starts and ends in the same spot but is split one frame different than the original
# Depending on how tightly your audio is edited this could cause small artifacts
# Video transitions are broken in a similar but slightly different way, so the same fix is applied to any video transitions with the END alignment too
#
# There are also some transitions incorrectly listed as CENTER that should have been END, probably due to rounding weirdness. This causes erroneous cutpointticks values that we have to properly detect and shift like an END transition
#
# Many transitions and clips also have an incorrect number of frames. We recalculate the frames from the rounded ticks then fix the rest of the variables based on the new frame count.
# This should more accurately reflect what was shown on the original timeline in Premiere Pro. Often the hover count of frames is wrong in PP 2021 if there are framerate rounding issues, because it rounds down values like 9.9999 length transitions to 9 frames, rather than the 10 if you manually count how many frames long the transition is.
#
# Because of the framerate rounding differences between 2019 and 2021 you'll also get flashes of rogue frames due to rounding errors, if your edits are tightly cut at the very start or end of a scene.
# We try to detect and fix those rounding differences mathematically, which can remove these artifacts but may also slightly alter the length of the film and the cuts between scenes may be a frame shorter to cut off the rogue frames.
#
# Transitions themselves may also be 1 frame shorter than the original 2021 versions due to rounding errors in Premiere Pro 2021. This isn't super common but does happen occasionally.
#
# This script will also upgrade all audio tracks to 5.1 dolby to make re-import easier, since it's easy to import stereo files into a dolby 5.1 timeline, but importing dolby 5.1 into a stereo timeline loses the extra tracks, no easy way to re-add them
#
# We also strip the file names and anonymize the file paths. If you have weird naming conventions you need to share you can set $cleanfilenames to $false.
#


# rather than rounding we're now capturing the frame count directly and using that instead 
# this is only a small fraction different than using framerate conversions but it seems slightly more accurate to how PP2021 calculates projects
function ConvertFrameCount
{
    param($framestring)

    #string to long
    $framestring = [System.Convert]::ToInt64($framestring)

    #fix the math (could be 1 frame wrong depending on the initial erroneous number that was rounded)
    # note that the subtraction is the opposite of doing this with framerates, since ticks go up as framerate goes down
    $framestring = [math]::Round($framestring + ((($fileticksperframe-$pproconst)/$pproconst)*$framestring))

    return [long]$framestring
}

function CalculatePproTicks
{
    param($framecount)

    # ticks are calculated off different framerate rounding depending on the version, typically need to be fixed
    # for example 2019 rounds framerates to 3 decimal places, 2021 uses 6 decimal places
    return ($framecount * $pproconst)
}

function CalcConvFramesFromTicks
{
    param($framestring)

    #string to long
    $framestring = [System.Convert]::ToInt64($framestring)

    # this is special for framecounts we need to round up
    [decimal]$unconvframes = $framestring/$fileticksperframe
    $framestring = [math]::Round($unconvframes + ((($fileticksperframe-$pproconst)/$pproconst)*$unconvframes))

    return [long]$framestring
}

function CalcConvFramesFromTicksUP
{
    param($framestring)

    #string to long
    $framestring = [System.Convert]::ToInt64($framestring)

    # this is special for framecounts we need to round up
    [decimal]$unconvframes = $framestring/$fileticksperframe
    $framestring = [math]::Ceiling($unconvframes + ((($fileticksperframe-$pproconst)/$pproconst)*$unconvframes))

    return [long]$framestring
}

function DecibelsToDecimal
{
    param($filtervalue)

    # convert from decibels to decimal
    $convertedgain = [System.Convert]::ToDecimal($filtervalue)
    $convertedgain = [math]::pow(10, ($convertedgain/20))
    $convertedgain = [math]::Round($convertedgain,6)

    return [decimal]$convertedgain
}

function ConstructExponentialFade($fgainresult)
{
    [long]$tmplng = 0

    $tmpstr = $filterleveltext.Replace('FILTERVALUE', "$fgainresult")

    if ($expfadein -eq 0 -and $expfadeout -eq 0)
    {
        # no fades to simulate, clear the placeholder
        $tmpstr = $tmpstr.Replace('FILTERKEYFRAMES', "")
    }
    elseif ($expfadein -ne 0 -and $expfadeout -ne 0)
    {
        $tmpstr2 = ""

        # simulate both in and out
        if ($expcacheinedge -ne -1 -and $expcacheininner -ne -1)
        {
            $tmpstr2 = $filterfadeintext

            # do the in side before concatenating because we use the same var names for the out and we need different positions obviously
            $tmplng = [math]::Round(($expcacheinedge + $expcacheininner)/2)

            $tmpstr2 = $tmpstr2.Replace('INNERPOS', "$expcacheininner")
            $tmpstr2 = $tmpstr2.Replace('HALFPOS', "$tmplng")
            $tmpstr2 = $tmpstr2.Replace('EDGEPOS', "$expcacheinedge")

            $global:expcacheinedge = -1
            $global:expcacheininner = -1
        }
                                            
        if ($expcacheoutedge -ne -1 -and $expcacheoutinner -ne -1)
        {
            $tmpstr2 = $tmpstr2 + $filterfadeouttext

            $tmplng = [math]::Round(($expcacheoutedge + $expcacheoutinner)/2)

            $tmpstr2 = $tmpstr2.Replace('INNERPOS', "$expcacheoutinner")
            $tmpstr2 = $tmpstr2.Replace('HALFPOS', "$tmplng")
            $tmpstr2 = $tmpstr2.Replace('EDGEPOS', "$expcacheoutedge")

            $global:expcacheoutedge = -1
            $global:expcacheoutinner = -1
        }

        # set the main volume for the clip, normally 1 but if we adjusted the gain it will be different
        $tmpstr2 = $tmpstr2.Replace('FILTERVALUE', "$fgainresult")
        $tmpstr = $tmpstr.Replace('FILTERKEYFRAMES', "$tmpstr2")

        $global:expfadein = 0
        $global:expfadeout = 0
    }
    elseif ($expfadein -ne 0)
    {
        # only in
        $tmpstr2 = ""

        if ($expcacheinedge -ne -1 -and $expcacheininner -ne -1)
        {
            $tmpstr2 = $filterfadeintext

            # do the in side before concatenating because we use the same var names for the out and we need different positions obviously
            $tmplng = [math]::Round(($expcacheinedge + $expcacheininner)/2)

            $tmpstr2 = $tmpstr2.Replace('INNERPOS', "$expcacheininner")
            $tmpstr2 = $tmpstr2.Replace('HALFPOS', "$tmplng")
            $tmpstr2 = $tmpstr2.Replace('EDGEPOS', "$expcacheinedge")

            $global:expcacheinedge = -1
            $global:expcacheininner = -1
        }

        # set the main volume for the clip, normally 1 but if we adjusted the gain it will be different
        $tmpstr2 = $tmpstr2.Replace('FILTERVALUE', "$fgainresult")
        $tmpstr = $tmpstr.Replace('FILTERKEYFRAMES', "$tmpstr2")

        $global:expfadein = 0
    }
    elseif ($expfadeout -ne 0)
    {
        # only out
        $tmpstr2 = ""

        if ($expcacheoutedge -ne -1 -and $expcacheoutinner -ne -1)
        {
            $tmpstr2 = $filterfadeouttext

            $tmplng = [math]::Round(($expcacheoutedge + $expcacheoutinner)/2)

            $tmpstr2 = $tmpstr2.Replace('INNERPOS', "$expcacheoutinner")
            $tmpstr2 = $tmpstr2.Replace('HALFPOS', "$tmplng")
            $tmpstr2 = $tmpstr2.Replace('EDGEPOS', "$expcacheoutedge")

            $global:expcacheoutedge = -1
            $global:expcacheoutinner = -1
        }

        # set the main volume for the clip, normally 1 but if we adjusted the gain it will be different
        $tmpstr2 = $tmpstr2.Replace('FILTERVALUE', "$fgainresult")
        $tmpstr = $tmpstr.Replace('FILTERKEYFRAMES', "$tmpstr2")

        $global:expfadeout = 0
    }

    return $tmpstr
}

# removed the keyframe xml from this filter, doesn't seem to be necessary so why complicate things
# we put the keyframes back if we want to simulate an exponential fade since that isn't imported properly
# if we only want a keyframe filter with no gain adjustment to the clip then the filter value gets set to 1
$global:filterleveltext = '						<filter>
							<effect>
								<name>Audio Levels</name>
								<effectid>audiolevels</effectid>
								<effectcategory>audiolevels</effectcategory>
								<effecttype>audiolevels</effecttype>
								<mediatype>audio</mediatype>
								<pproBypass>false</pproBypass>
								<parameter authoringApp="PremierePro">
									<parameterid>level</parameterid>
									<name>Level</name>
									<valuemin>0</valuemin>
									<valuemax>3.98109</valuemax>
									<value>FILTERVALUE</value>
									FILTERKEYFRAMES
								</parameter>
							</effect>
						</filter>
'

# we go from -16db at the edge to -12db at half to 0db at inner to give the illusion of a curve more similar to exponential fade
# filtervalue will generally be 1 unless there's a gain applied to the clip that we're compensating for with levels
$global:filterfadeouttext = '<keyframe>
										<when>INNERPOS</when>
										<value>FILTERVALUE</value>
									</keyframe>
									<keyframe>
										<when>HALFPOS</when>
										<value>0.251189</value>
									</keyframe>
									<keyframe>
										<when>EDGEPOS</when>
										<value>0.158489</value>
									</keyframe>'

$global:filterfadeintext = '<keyframe>
										<when>EDGEPOS</when>
										<value>0.158489</value>
									</keyframe>
									<keyframe>
										<when>HALFPOS</when>
										<value>0.251189</value>
									</keyframe>
									<keyframe>
										<when>INNERPOS</when>
										<value>FILTERVALUE</value>
									</keyframe>'


# need this for easy reworking of audio
$dolby5output = '				<outputs>
					<group>
						<index>1</index>
						<numchannels>1</numchannels>
						<downmix>0</downmix>
						<channel>
							<index>1</index>
						</channel>
					</group>
					<group>
						<index>2</index>
						<numchannels>1</numchannels>
						<downmix>0</downmix>
						<channel>
							<index>2</index>
						</channel>
					</group>
					<group>
						<index>3</index>
						<numchannels>1</numchannels>
						<downmix>0</downmix>
						<channel>
							<index>3</index>
						</channel>
					</group>
					<group>
						<index>4</index>
						<numchannels>1</numchannels>
						<downmix>0</downmix>
						<channel>
							<index>4</index>
						</channel>
					</group>
					<group>
						<index>5</index>
						<numchannels>1</numchannels>
						<downmix>0</downmix>
						<channel>
							<index>5</index>
						</channel>
					</group>
					<group>
						<index>6</index>
						<numchannels>1</numchannels>
						<downmix>0</downmix>
						<channel>
							<index>6</index>
						</channel>
					</group>
				</outputs>'


# --------------------------
# VARIABLES 
# --------------------------

# this path has subfolders e.g. M:\movie-edits-exports\premiere2019 that contain xml files to be processed, this script will create these and the subfolders if you run it and they don't exist
$basepath = "M:\movie-edits-exports\"

# this is purely for organization if you want to categorize your edits by premiere version, it currently has no impact on how the data is processed
$premiere_versions = @("2019","2021","2023","2024")

# It's a good idea to leave this as a fake path so that when you reimport a video into your project, Premiere will ask you to find it, and give you errors if there are any mismatch issues
# If you pick a file that's a different audio type it won't let you select it, however you can cancel the file selection process and right-click the clip in the project, using the Replace Footage to select your video with a different audio channel count
$cleanpath = 'file://localhost/C%3a/source-films-tv/'

# If you want a release group name or some other text inserted in the output file name after the resolution
# Start this text with a . so it looks correct!
$release_name = ''

# --------------------------
# SETTINGS
# --------------------------

# toggle off if you don't want it to clean the file names for you
# it takes something like Zombieland.2009.1080p.BluRay.x264-RELEASEGROUP.mp4 and turns it into Zombieland.2009.1080p.youredit.mp4
# or Game.of.Thrones.S04E09.The.Watchers.On.the.Wall.1080p.DD5.1.H.264.mp4 into Game.of.Thrones.S04E09.1080p.youredit.mp4
$cleanfilenames = $true 

# this is a subtoggle of cleanfilenames, it creates a version database and adds the version number to your output file, incrementing every time you run this script on the same cleaned file title
$add_version = $true

# Premiere breaks exponential fades and converts them to 0db crossfades, which import as constant gain
# if you only use constant power and exponential fade then we can assume any constant gain are actually exponential fades and convert accordingly
# if you use constant gain transitions then disable this conversion
# WARNING: because of the heavy regex processing involved this is significantly slower - up to a couple minutes per file - especially with projects that have tons of cuts/xml
$convert_0db_to_fade = $true

# toggle off if you don't want it to reprocess audio into 5.1 format for better import compatibility, 
# or if you're reprocessing a file with audio channels on different tracks already so each channel isn't turned into 6 channels
# think of this as a lossy process, you really shouldn't be re-exporting an imported fixed xml, go back and edit the original project
$simulate_51_audio = $true 

# this is a subtoggle of simulate_51_audio - making this true puts each audio channel on a separate track, if AME is crashing this may fix it
# currently with this off, you can render in Premiere Pro, but if you try to Queue and render in AME it will fail, not sure what AME doesn't like, seems to be something with the audio channels
$split_audio_channels = $true 

# this re-numbers the videos so that their ID is consistent across different episodes and even projects. Once a source video is seen it will always keep the same sequentally assigned number.
$cleanvidid = $true

# this corresponds to an index in the $fpsarr below. Normally you want to leave this at -1 to let the script auto-detect the nearest framerate, but if you need to force the framerate for legacy reasons set this. 
# For example 0 = 23.976fps, 1 = 24fps, etc.
$forceframerate = -1

# This will recalculate the split in any transition to keep it positioned the same but as a center rather than a start/end,
# turn it off if you want transition positioning to reflect your original project.  Keeping this on allows for better compatibility
# if you're copying from one sequence to another, as sometimes Premiere will re-center transitions when they're copied between sequences.
# So if we center them beforehand they'll look different on the timeline but ensure the effect stays consistent.
$centertransitions = $true

# --------------------------
# CONSTANTS (don't touch)
# --------------------------
$targetext = "*.xml" 
$companionext = ".ale"

# going down as framerate goes up, ticks seem to be how long each frame is, thus a higher number of frames per second equates to shorter frames (smaller number)
[long]$ppro23976fps = 10594584000 
[long]$ppro24fps = 10584000000 
[long]$ppro25fps = 10160640000 
[long]$ppro2997fps = 8475667200
[long]$ppro48fps = 5292000000
[long]$ppro50fps = 5080320000
[long]$ppro5994fps = 4237833600
[long]$ppro60fps = 4233600000 

# To add a new framerate/ticks ratio do this:
# 60/newfps = fraction
# fraction (rounded to 4 digits after decimal) * 4233600000 = new ticks value
# don't forget to also add the appropriate values to the timebase and ntsc arrays as well
# the appropriate values are explained further below



# 10584000000 / 4233600000 = 2.5
# 60/2.5 = 24fps

# 60fps/25fps = 2.4 * 4233600000 = 10,160,640,000
# 60fps/23.976fps = 2.502502502502503 - doesn't work, probably rounded
# 10594584000 / 4233600000 = 2.5025 - yeah, looks like 4 digits

# let's shove them in an array for ease of use
$fpsarr = @($ppro23976fps, $ppro24fps, $ppro25fps, $ppro2997fps, $ppro48fps, $ppro50fps, $ppro5994fps, $ppro60fps)
# need timebase to go along with the framerate, seems to be a 2 digit rounded, so 23.976 is also 24
# IMPORTANT: keep this array the same length and order as $fpsarr!
$timebasearr = @(24,24,25,30,48,50,60,60)
# the ntsc value is what turns the 24 timebase into a 23.976 framerate
$ntscarr = @('TRUE','FALSE','FALSE','TRUE','FALSE','FALSE','TRUE','FALSE')

# initializing this as 24fps, but it's going to change for each file processed
# the timebase goes with the sequence, not the video file, so it will be consistent regardless of the different clip framerates, like ticks are
[long]$pproconst = $fpsarr[1]
$pprotimebase = $timebasearr[1]
$pprontsc = $ntscarr[1]

$fixclipconst = 1
$fixtransconst = 2

$global:expfadesoloconst = 1
$global:expfadesharedconst = 2

$filtergainconst = 1
$filterlevelconst = 2
# --------------------------
# --------------------------

$versions_ht = @{}
$versiondb = $basepath + "versiondb.txt"
$srcvid_ht = @{}
$srcviddb = $basepath + "sourcevideodb.txt"
$srcvidmax = 0

# just to avoid any unexpected logic errors from suboptions being on while the main option is off
if ($cleanfilenames -ne $true)
{
    $add_version = $false
}

if ($simulate_51_audio -ne $true)
{
    $split_audio_channels = $false
}

# load our versions database
if ($add_version -eq $true)
{
    if ([System.IO.File]::Exists($versiondb))
    {
        [System.IO.File]::ReadLines($versiondb) | ForEach-Object {
            $linedata_orig = $_

            $tmparr = $linedata_orig.Split(",")

            if ($tmparr.Length -eq 2)
            {
                $versions_ht.add($tmparr[0], [System.Convert]::ToInt64($tmparr[1]))
            }
        }
    }
    else
    {
        # Write-Output "Version database does not exist, creating: '$versiondb'"
    }
}

if ($cleanvidid -eq $true)
{
    if ([System.IO.File]::Exists($srcviddb))
    {
        [System.IO.File]::ReadLines($srcviddb) | ForEach-Object {
            $linedata_orig = $_

            $tmparr = $linedata_orig.Split(",")

            if ($tmparr.Length -eq 2)
            {
                $tmpint = [System.Convert]::ToInt64($tmparr[1])

                # track the max value we've used so we don't overlap
                if ($tmpint -gt $srcvidmax)
                {
                    $srcvidmax = $tmpint
                }

                $srcvid_ht.add($tmparr[0], $tmpint)
            }
        }
    }
    else
    {
        # Write-Output "Source video ID database does not exist, creating: '$srcviddb'"
    }
}

# start processing xml files
foreach ($pversion in $premiere_versions) 
{
    $srcpath = $basepath + "premiere" + $pversion + "\" 
    $destpath = $srcpath + "output\"
    $finpath = $srcpath + "done\"

    $silenceoutput = New-Item -ItemType Directory -Force -Path $destpath
    $silenceoutput = New-Item -ItemType Directory -Force -Path $finpath

    

    Get-ChildItem -path ( $srcpath + $targetext ) | foreach {
        $origname1 = $_.Name
        $origname1NOext = $_.BaseName
        #$origname1_normal = $origname1.Replace('(',' ').Replace(')',' ').Replace('[',' ').Replace(']',' ')
        $origname1_normal = $origname1 -replace '[\[\]\(\)_,]', ' ' # stage 1, replace crap with spaces
        $origname1_normal = $origname1_normal -replace '\s+', '.' # stage 2, replace multiple spaces in a row with one period
        $ismatch = $origname1_normal -match '(.+)(1080p|720p)'
        $cleanname = ""
        $newname = ""
        $newtitle = ""
        $print_audio_outputs = $false
        $firstdolbytrack = ""
        $caching_dolby_track = $false
        $extra_clip_autonumber = 10000 # nice high starting value to avoid conflicting with existing clips
        $nuking_track = $false
        $link_count = 0
        $link_expected = 0
        $create_linkcache = $false
        $link_template = ""
        $link_latest = ""
        $link_ht = @{}
        $latest_clip = ""
        $tmpht = ""
        $physical_track_count = -1 # zero based

        $cutpointshift = $pproconst # how many ticks to subtract from end cutpointticks, this is adjusted by a function later to be more accurate
        $donotmove = $true

        $wholefile = ""

        if ($ismatch -eq $true)
        {
            Write-Output "Processing '$origname1'."
            
            [long]$fileversion = 1

            $origext = $_.Extension

            if ($cleanfilenames -eq $true)
            {
                # name plus resolution only
                $cleanname = $Matches.1 + $Matches.2
        
                # try a secondary more detailed match for tv episodes and movies with a year to clean up DTS, etc
                if ($origname1_normal -match '(.+)(s[0-9]{2}e[0-9]{2}-[0-9]{2})(.*)(1080p|720p)') 
                {
                    # this is for spanning multiple episodes example: Californication.S06E04-05.1080p.xml
                    $cleanname = $Matches.1 + $Matches.2 + "." + $Matches.4
                }
                elseif ($origname1_normal -match '(.+)(s[0-9]{2}e[0-9]{2})(.*)(1080p|720p)') 
                {
                    # this works for both normal tv episodes (Californication.S06E02.1080p.xml) as well as episodes spanning two seasons (Californication.S02E12-S04E01.1080p.xml) because the first wildcard acts greedy
                    $cleanname = $Matches.1 + $Matches.2 + "." + $Matches.4
                }
                elseif ($origname1_normal -match '(.+)([0-9]{4}\.e[0-9]{2})(.*)(1080p|720p)') 
                {
                    # look for movie series titles that look like this: Resident.Evil.2004-2016.E01.1080p.xml
                    $cleanname = $Matches.1 + $Matches.2 + "." + $Matches.4
                }
                elseif ($origname1_normal -match '(.+)([0-9]{4})(.*)(1080p|720p)') 
                {
                    # movies with year in title
                    $cleanname = $Matches.1 + $Matches.2 + "." + $Matches.4
                }

                $newtitle = $cleanname + $release_name

                if ($add_version -eq $true)
                {
                    # retrieve last version if it exists
                    $tmpver = $versions_ht[$newtitle]

                    if ($tmpver.Length -gt 0)
                    {
                        $fileversion = [System.Convert]::ToInt64($tmpver)
                        $fileversion++
                    }

                    $versions_ht[$newtitle] = $fileversion

                    # append proper version number to the names
                    $newtitle += ".ver" + $fileversion
                }
                
                $newname = $destpath + $newtitle + $origext
            }
            else
            {
                $cleanname = $origname1_normal
                $newname = $destpath + $cleanname
                $newtitle = $cleanname
            }
            
            $firstname = $true # first <name> tag is a project name, not necessarily a file name, we're going to fix those differently just to be sure
            $isstrip = 0

            # before we create our new file we also need the ALE (avid log exchange) file
            [decimal]$fileframerate = -1
            [long]$fileticksperframe = -1
           
           # This method is close, but lets see if we can use a different method with the exact rate for better accuracy with 2021's 6 digit decimal

           #if ([System.IO.File]::Exists($srcpath + $origname1NOext + $companionext)) {
           ##Get-ChildItem -path ( $srcpath + $origname1NOext + $companionext ) | foreach {
           #    [System.IO.File]::ReadLines($srcpath + $origname1NOext + $companionext) | ForEach-Object {
           #        $linedata_orig = $_
           #
           #        # we want this row:
           #        #FPS	23.976
           #
           #        if ($linedata_orig.Length -ge 3 -and $linedata_orig.Substring(0,3) -eq "FPS")
           #        {
           #            # if the fileframerate isn't 24 we're going to recalculate every value to make sure we don't have buggy Adobe Premiere output
           #            $fileframerate = [System.Convert]::ToDecimal($linedata_orig.Substring(4))
           #        }
           #
           #    }
           #}
            
            $trackactive = $false
            $transitionactive = $false
            [long]$tmpin = -1
            [long]$tmpout = -1
            [long]$tmpticksin = -1
            [long]$tmpticksout = -1
            $tmpaligned = $false
            $tmpstr = ""
            $currentfixtype = -1
            $isproject = $false

            # instead of the previous methods of processing each line we're going to grab the entire file so we can pre-process it in certain ways before doing the conversion pass
            $wholefile = [System.IO.File]::ReadAllText($srcpath + $origname1)

            # standardize newlines
            $wholefile = $wholefile -replace "`r`n", "`n"
            $wholefile = $wholefile -replace "`r", "`n"

            $wholefile_lines = $wholefile -split "`n"

            # we're going to pre-read the first part of the file to grab the data to calculate the framerate
            #[System.IO.File]::ReadLines($srcpath + $origname1) | ForEach-Object {
            ForEach ($linedata_orig in $wholefile_lines) {

                if ($fileframerate -eq -1) # can't break out of a foreach-object, just read the rest of the lines quickly
                {
                    #$linedata_orig = $_

                    switch -regex ($linedata_orig)
                    {
                        '(<clipitem )' { 
                            # we can use clips without transitions attached to easily calculate the exact framerate
                            $trackactive = $true
                            $transitionactive = $false
                            $tmpin = -1
                            $tmpout = -1
                            $tmpticksin = -1
                            $tmpticksout = -1
                            $tmpaligned = $true # set clips aligned automatically
                            break
                        }

                        '(</clipitem>)' { 
                            # Reset everything and wait for next clip if we didn't calculate the data we needed from this one
                            $trackactive = $false
                            $transitionactive = $false
                            $tmpin = -1
                            $tmpout = -1
                            $tmpticksin = -1
                            $tmpticksout = -1
                            $tmpaligned = $false
                            break
                        }

                        '(<transitionitem>)' { 
                            # we can also try to use audio transitions to calculate the framerate but the process is slightly different
                            $trackactive = $false
                            $transitionactive = $true
                            $tmpin = -1
                            $tmpout = -1
                            $tmpticksin = -1
                            $tmpticksout = -1
                            $tmpaligned = $false
                            break
                        }

                        '(</transitionitem>)' { 
                            # Reset everything and wait for next clip if we didn't calculate the data we needed from this one
                            $trackactive = $false
                            $transitionactive = $false
                            $tmpin = -1
                            $tmpout = -1
                            $tmpticksin = -1
                            $tmpticksout = -1
                            $tmpaligned = $false
                            break
                        }
                    

                        '(<in>)(.*)(</in>)' { 
                            if ($trackactive -eq $true)
                            {
                                $tmpstr = $linedata_orig -replace '(<in>)(.*)(</in>)', "`$2"
                                $tmpin = [System.Convert]::ToInt64($tmpstr)
                            }
                            break
                        }

                        '(<out>)(.*)(</out>)' { 
                            if ($trackactive -eq $true)
                            {
                                $tmpstr = $linedata_orig -replace '(<out>)(.*)(</out>)', "`$2"
                                $tmpout = [System.Convert]::ToInt64($tmpstr)
                            }    
                            break
                        }

                        '(<pproTicksIn>)(.*)(</pproTicksIn>)' { 
                            if ($trackactive -eq $true -or $transitionactive -eq $true)
                            {
                                $tmpstr = $linedata_orig -replace '(<pproTicksIn>)(.*)(</pproTicksIn>)', "`$2"
                                $tmpticksin = [System.Convert]::ToInt64($tmpstr)
                            }   
                            break
                        }

                        '(<pproTicksOut>)(.*)(</pproTicksOut>)' { 
                            if ($trackactive -eq $true -or $transitionactive -eq $true)
                            {
                                $tmpstr = $linedata_orig -replace '(<pproTicksOut>)(.*)(</pproTicksOut>)', "`$2"
                                $tmpticksout = [System.Convert]::ToInt64($tmpstr)
                            }   
                            break
                        }

                        '(<start>)(.*)(</start>)' { 
                            if ($trackactive -eq $true)
                            {
                                $tmpstr = $linedata_orig -replace '(<start>)(.*)(</start>)', "`$2"
                                if ([System.Convert]::ToInt64($tmpstr) -eq -1)
                                {
                                    # There's some sort of padding involved in clips that have transitions attached, if this has a transition on either side we're ignoring it
                                    $trackactive = $false
                                    $transitionactive = $false
                                    $tmpin = -1
                                    $tmpout = -1
                                    $tmpticksin = -1
                                    $tmpticksout = -1
                                    $tmpaligned = $false
                                }
                            }   

                            if ($transitionactive -eq $true)
                            {
                                # for transitions START is our IN tag
                                $tmpstr = $linedata_orig -replace '(<start>)(.*)(</start>)', "`$2"
                                $tmpin = [System.Convert]::ToInt64($tmpstr)
                            }
                            break
                        }

                        '(<end>)(.*)(</end>)' { 
                            if ($trackactive -eq $true)
                            {
                                $tmpstr = $linedata_orig -replace '(<end>)(.*)(</end>)', "`$2"
                                if ([System.Convert]::ToInt64($tmpstr) -eq -1)
                                {
                                    # There's some sort of padding involved in clips that have transitions attached, if this has a transition on either side we're ignoring it
                                    $trackactive = $false
                                    $transitionactive = $false
                                    $tmpin = -1
                                    $tmpout = -1
                                    $tmpticksin = -1
                                    $tmpticksout = -1
                                    $tmpaligned = $false
                                }
                            }   

                            if ($transitionactive -eq $true)
                            {
                                # for transitions END is our OUT tag
                                $tmpstr = $linedata_orig -replace '(<end>)(.*)(</end>)', "`$2"
                                $tmpout = [System.Convert]::ToInt64($tmpstr)
                            }
                            break
                        }

                        '(<alignment>)(.*)(</alignment>)' { 

                            if ($transitionactive -eq $true)
                            {
                                $tmpalign = $linedata_orig -replace '(<alignment>)(.*)(</alignment>)', "`$2"
                                $tmpalign = $tmpalign.ToLower().Trim()

                                if ($tmpalign -eq "center")
                                {
                                    $tmpaligned = $true # center is the only alignment type that seems reliable, the rest are either slightly or majorly off
                                }
                                
                            }
                            break

                        }

                        '(<project>)' { 
                            # we only want sequences, not projects, warn the user
                            $isproject = $true
                            break
                        }

                    }

                    if ($tmpin -ne -1 -and $tmpout -ne -1 -and $tmpticksin -ne -1 -and $tmpticksout -ne -1 -and $tmpaligned -eq $true)
                    {
                        if ($trackactive -eq $true)
                        {
                            # got all our variables, calculate it
                            $tmpout -= $tmpin
                            $tmpticksout -= $tmpticksin

                            $tmpticksout = $tmpticksout/$tmpout

                            $fileticksperframe = $tmpticksout
                            $fileframerate = ($ppro24fps/$tmpticksout)*24

                            
                        }
                        else
                        {
                            # using transitions we need a bit of fuzzy logic because there's a rounding error where the start tag sometimes needs to have 1 subtracted to be the correct frame count
                            # do it both ways and compare the result, if one is a decimal value then we know the other is the correct one.  If both are whole numbers toss them out and start again.
                            $tmpout -= $tmpin
                            $tmpticksout -= $tmpticksin

                            [decimal]$test1 = $tmpticksout/$tmpout

                            $tmpout++

                            [decimal]$test2 = $tmpticksout/$tmpout

                            if ((($test1 % 1) -eq 0) -and (($test2 % 1) -ne 0))
                            {
                                # test1 is correct value
                                $fileticksperframe = $test1
                                $fileframerate = ($ppro24fps/$test1)*24
                            } 
                            elseif ((($test2 % 1) -eq 0) -and (($test1 % 1) -ne 0))
                            {
                                # test2 is correct value
                                $fileticksperframe = $test2
                                $fileframerate = ($ppro24fps/$test2)*24
                            }
                            else
                            {
                                $fileticksperframe = -1
                                $fileframerate = -1
                            }

                            # if neither of those tests worked we either have two decimals or two whole numbers, toss them out and calculate the next one

                        }
                        
                        if ($fileframerate -ne -1)
                        {
                            # seems the clip division is rounded to 13 digits, round the transition math to the same length to keep everything standardized
                            $fileframerate = [math]::Round($fileframerate,13)
                        }

                        $trackactive = $false
                        $tmpin = -1
                        $tmpout = -1
                        $tmpticksin = -1
                        $tmpticksout = -1
                        $tmpaligned = $false
                    }
                }
            }

            if ($isproject -eq $false -and $fileframerate -ne -1)
            {
                # assign the target ticks value based on the closest match to the original project
                [long]$lastdiff = -1

                for ($i=0;$i -lt $fpsarr.Length;$i++)
                {
                    [long]$fpsdiff = 0

                    if ($fileticksperframe -gt $fpsarr[$i])
                    {
                        $fpsdiff = $fileticksperframe - $fpsarr[$i]
                    }
                    else
                    {
                        $fpsdiff = $fpsarr[$i] - $fileticksperframe
                    }

                    if ($lastdiff -eq -1 -or $fpsdiff -lt $lastdiff)
                    {
                        $pproconst = $fpsarr[$i]
                        $pprotimebase = $timebasearr[$i]
                        $pprontsc = $ntscarr[$i]
                        $lastdiff = $fpsdiff
                    }
                }

                if ($forceframerate -gt -1)
                {
                    $pproconst = $fpsarr[$forceframerate]
                    $pprotimebase = $timebasearr[$forceframerate]
                    $pprontsc = $ntscarr[$forceframerate]
                }

                #Write-Output "Detected rate: '$fileticksperframe','$fileframerate' - Using conversion values '$pproconst','$pprotimebase'"

                if ($cleanvidid -eq $true)
                {
                    # this is basically the same code from list-project-file-indexes, repurposed to be more automated rather than something the user has to run themselves during calibration
                    [long]$masterclip_id = -1
                    $masterclip_ht = @{}

                    ForEach ($linedata_orig in $wholefile_lines) {

                        switch -regex ($linedata_orig)
                        {
                            '(<clipitem )' { 
                                $masterclip_id = -1
                                break
                            }

                            '(</clipitem>)' { 
                                $masterclip_id = -1
                                break
                            }

                            '(<masterclipid>)' { 
                                $tmpstr = $linedata_orig -replace '(<masterclipid>)(.*)(</masterclipid>)', "`$2"
                                $tmpstr = $tmpstr.ToLower().Trim()

                                # make the result just the number
                                $masterclip_id = $tmpstr.Replace("masterclip-", "")
                                break
                            }

                            # found a <name> tag
                            '(<name>)(.*)(</name>)' { 
                                if ($masterclip_id -ne -1)
                                {
                                    $tmpstr = $linedata_orig -replace '(<name>)(.*)(</name>)', "`$2"
                                    $tmpstr = $tmpstr.Trim()

                                    if ($masterclip_ht.ContainsKey($masterclip_id) -eq $false)
                                    {
                                        $masterclip_ht.add($masterclip_id, $tmpstr)
                                    }

                                    # don't want to accidentally pick up a name from a filter or something
                                    $masterclip_id = -1
                                }
                                break
                            }

                            '(<project>)' { 
                                # we only want sequences, not projects, warn the user
                                $isproject = $true
                                break
                            }

                        }

                    }

                    foreach ($tmppair in $masterclip_ht.GetEnumerator())
                    {
                        #$tmpstr = "file: " + $tmppair.Value + ", index: " + $tmppair.Name

                        # normalize name first
                        $tmpstr = $tmppair.Value

                        $tmpstr = $tmpstr -replace '[\[\]\(\)_,]', ' ' # stage 1, replace crap with spaces
                        $tmpstr = $tmpstr -replace '\s+', '.' # stage 2, replace multiple spaces in a row with one period

                        if ($tmpstr -match '(.+)(s[0-9]{2}e[0-9]{2})(.*)(1080p|720p)(.*?)(.[^.]+)$') # try a secondary more detailed match for tv episodes
                        {
                            $tmpstr = $Matches.1 + $Matches.2 + "." + $Matches.4 + $Matches.6
                        }
                        elseif ($tmpstr -match '(.+)([0-9]{4})(.*)(1080p|720p)(.*?)(.[^.]+)$') # try a secondary more detailed match for movies with year to clean up DTS, etc
                        {
                            $tmpstr = $Matches.1 + $Matches.2 + "." + $Matches.4 + $Matches.6
                        }
                        elseif ($tmpstr -match '(.+)(1080p|720p)')
                        {
                            $tmpstr = $tmpstr -replace '(.+)(1080p|720p)(.*?)(.[^.]+)$', '$1$2$4'
                        }

                        # now that we know all the videos in this xml file we want to cross-reference them to see if it's one we've used before - if so retrieve the index
                        # if it isn't something we've used, assign a new index for it

                        $tmpsrcvid = $srcvid_ht[$tmpstr]
                        

                        if ($tmpsrcvid.Length -gt 0)
                        {
                            $tmpsrcvid = [System.Convert]::ToInt64($tmpsrcvid)
                        }
                        else
                        {
                            $srcvidmax = $srcvidmax + 1
                            $tmpsrcvid = $srcvidmax
                            $srcvid_ht[$tmpstr] = $tmpsrcvid
                        }

                        # then we need to replace all references to it, these come in two formats:
                        #<masterclipid>masterclip-129</masterclipid>
                        #<file id="file-129"/>

                        $tmpstr = 'masterclip-' + $tmppair.Name + '<'
                        $tmpstr2 = 'masterclipbestedit-' + $tmpsrcvid + '<' # note we're putting bestedit in the name to avoid any bugs where we replace a value and then replace it again with a new one. At the end of the replaces we'll switch this back.
                        $wholefile = $wholefile.Replace($tmpstr, $tmpstr2) # note that we're reading from wholefile_lines and replacing values in wholefile since that's what later script sections use

                        $tmpstr = 'file-' + $tmppair.Name + '"'
                        $tmpstr2 = 'filebestedit-' + $tmpsrcvid + '"'
                        $wholefile = $wholefile.Replace($tmpstr, $tmpstr2)
                    }

                    $wholefile = $wholefile.Replace('masterclipbestedit-', 'masterclip-')
                    $wholefile = $wholefile.Replace('filebestedit-', 'file-')
                }

                # The speed effect is hard to fix in the original format, we're going to regex it to make some changes and insert our own tags that make it easier to reprocess accurately
                # $firstdolbytrack = $firstdolbytrack -replace '(?ms)(<link>((?!<link>).)*?<mediatype>video</mediatype>.*?</link>)', ""
                while ($wholefile -match '(?ms)(<clipitem ((?!<clipitem ).)*?<effectid>timeremap</effectid>.*?<parameterid>graphdict</parameterid>.*?</clipitem>)')
                {
                    $speedclip = $wholefile -replace '(?ms)(.*)(<clipitem ((?!<clipitem ).)*?<effectid>timeremap</effectid>.*?<parameterid>graphdict</parameterid>.*?</clipitem>)(.*)', "`$2"
                    $speedclip_orig = $speedclip

                    # grab the multiplier for the speed
                    #
                    #<parameter authoringApp="PremierePro">
					#	<parameterid>speed</parameterid>
					#	<name>speed</name>
					#	<valuemin>-100000</valuemin>
					#	<valuemax>100000</valuemax>
					#	<value>92.3592</value>
					#</parameter>
                    $speedparam = $speedclip -replace '(?ms)(.*)(<parameter ((?!<parameter ).)*?<parameterid>speed</parameterid>.*?<value>(.*?)</value>.*?</parameter>)(.*)', "`$4"
                    
                    # Now attach the speed as a new tag right after the start of the clip, so we have it when we're processing the other variables
                    # <clipitem id="clipitem-329">
                    $speedclip = $speedclip -replace '(?ms)(<clipitem ((?!>).)*?>)', "`$1`n`t`t`t`t`t`t<besteditspeed>$speedparam</besteditspeed>"

                    # get rid of graphdict keyframes
                    # it doesn't seem like we actually need these, we just need to set the in/out tags on the clip properly, which in this case is actually using an incorrect value based on the speed percentage (in / speedclip percentage)
                    $speedclip = $speedclip -replace '(?ms)(<parameter ((?!<parameter ).)*?<parameterid>graphdict</parameterid>.*?</parameter>)', ""

                    # put our reprocessed speed clip back into the place of the original
                    #$wholefile = $wholefile -replace $speedclip_orig, $speedclip
                    $wholefile = $wholefile.Replace($speedclip_orig, $speedclip)
                }

                if ($convert_0db_to_fade -eq $true)
                {
                    # if we want to convert the 0db to exponential fade keyframes then we need to pre-insert some extra xml to make processing the clips easier
                    # <effectid>KGAudioTransCrossFade0dB</effectid>

                    # This regex 0 is matching one clip, two transitions, and one clip, and we don't want it to match in that situation, only with one transition in the middle
                    # the first wildcard is eating the entire first transition, forbid that
                    # '(?ms)(<clipitem ((?!<clipitem ).)*?<effectid>KGAudioTransCrossFade0dB</effectid>((?!<transitionitem>).)*?</clipitem>)'
                    # once we fix that one the second also starts eating double transitions, first it does one then the next loop it grabs the second one, apply the same fix to it
                    # add a preemptive fix to the last regex too so it doesn't grab multiple clips in an edge case

                    $fade_regex = @('(?ms)(<clipitem ((?!<clipitem )(?!</transitionitem>).)*?<effectid>KGAudioTransCrossFade0dB</effectid>((?!<transitionitem>).)*?</clipitem>)', # this regex should match one clip, one transition, and another clip
                                    '(?ms)(<clipitem ((?!<clipitem )(?!</transitionitem>).)*?<effectid>KGAudioTransCrossFade0dB</effectid>((?!<transitionitem>).)*?</transitionitem>)', # this regex should match one clip, one transition
                                    '(?ms)(<transitionitem>((?!</transitionitem>).)*?<effectid>KGAudioTransCrossFade0dB</effectid>((?!<transitionitem>)(?!</clipitem>).)*?</clipitem>)') # this regex should match one transition, one clip

                    $fade_regex2 = @('(?ms)(.*)(<clipitem ((?!<clipitem )(?!</transitionitem>).)*?<effectid>KGAudioTransCrossFade0dB</effectid>((?!<transitionitem>).)*?</clipitem>)(.*)',
                                    '(?ms)(.*)(<clipitem ((?!<clipitem )(?!</transitionitem>).)*?<effectid>KGAudioTransCrossFade0dB</effectid>((?!<transitionitem>).)*?</transitionitem>)(.*)',
                                    '(?ms)(.*)(<transitionitem>((?!</transitionitem>).)*?<effectid>KGAudioTransCrossFade0dB</effectid>((?!<transitionitem>)(?!</clipitem>).)*?</clipitem>)(.*)')

                    for ($i=0;$i -lt $fade_regex.Length;$i++)
                    {
                        while ($wholefile -match $fade_regex[$i])
                        {
                            $speedclip = $wholefile -replace $fade_regex2[$i], "`$2"
                            $speedclip_orig = $speedclip

                            #Write-Output "speedmatch regex '$i', '$speedclip'"

                            $tmpcenter = $false
                            [long]$tmplng = 0
                            [long]$tmplng2 = 0
                            [long]$tmplng3 = 0

                            # for our first regex where we get two clips with a transition in the center we want to check if both are attached to the transition
                            # that way we can fade the zero side properly if it's a start or end transition
                            # this isn't completely necessary but if you have a start/end that has a loud sound on the zero side it will bleed through the fade if we don't add keyframes to silence it half way through the fade.
                            if ($i -eq 0)
                            {
                                # get END from first clip, and START from second clip
                                $tmplng = ($speedclip -replace '(?ms)(.*)(<clipitem )(.*?)<end>(.*?)</end>(.*?)(<transitionitem>)(.*)', "`$4").Trim()
                                $tmplng2 = ($speedclip -replace '(?ms)(.*)(</transitionitem>)(.*?)<start>(.*?)</start>(.*)', "`$4").Trim()

                                if ($tmplng -eq -1 -and $tmplng2 -eq -1)
                                {
                                    $tmpcenter = $true
                                }
                            }

                            # grab the alignment
                            $speedparam = $speedclip -replace '(?ms)(.*)(<transitionitem>((?!</transitionitem>).)*?<alignment>(.*?)</alignment>.*?</transitionitem>)(.*)', "`$4"
                            $speedparam = $speedparam.ToLower().Trim()

                            # grab the ticks, we're going to use these in our new tag to simplify things later
                            $tmplng = ($speedclip -replace '(?ms)(.*)(<transitionitem>((?!</transitionitem>).)*?<pproTicksIn>(.*?)</pproTicksIn>.*?</transitionitem>)(.*)', "`$4").Trim()
                            $tmplng2 = ($speedclip -replace '(?ms)(.*)(<transitionitem>((?!</transitionitem>).)*?<pproTicksOut>(.*?)</pproTicksOut>.*?</transitionitem>)(.*)', "`$4").Trim()
                            $tmplng3 = ($speedclip -replace '(?ms)(.*)(<transitionitem>((?!</transitionitem>).)*?<cutPointTicks>(.*?)</cutPointTicks>.*?</transitionitem>)(.*)', "`$4").Trim()

                            # for the purposes of this fade simulation we don't care where the cutpointticks fall, the underlying clips always reach the entire length of the transition whether it's start, end, or center aligned
                            # so for an exponential fade we want to target the start of our fade at the inner edge of the transition and fade half the length between ticksin and ticksout
                            # if the transition is not double sided then we fade the entire length between ticksin and ticksout
                            
                            # transition length in ticks
                            $tmplng3 = $tmplng2 - $tmplng

                            if ($tmpcenter -eq $true)
                            {
                                # for center fade out half way through
                                $tmplng3 = $tmplng3 / 2
                                # make it negative so we can detect it later
                                $tmplng3 = -$tmplng3
                            }

                            # Now attach the indicator tags right after the start of each clip, so we have it when we're processing the other variables
                            # <clipitem id="clipitem-329">
                            switch ($speedparam)
                            {
                                'center' {
                                    $speedclip = $speedclip -replace '(?ms)(<clipitem ((?!>).)*?>)(.*)(<clipitem ((?!>).)*?>)', "`$1`n`t`t`t`t`t`t<besteditfadeout>$tmplng3</besteditfadeout>`$3`$4`n`t`t`t`t`t`t<besteditfadein>$tmplng3</besteditfadein>"
                                    break
                                }

                                'end-black' {
                                    if ($tmpcenter -eq $true)
                                    {
                                        # this should never be true, but just in case...
                                        Write-Output "Warning: End-Black conversion anomaly at '$tmplng'"
                                        $speedclip = $speedclip -replace '(?ms)(<clipitem ((?!>).)*?>)(.*)(<clipitem ((?!>).)*?>)', "`$1`n`t`t`t`t`t`t<besteditfadeout>$tmplng3</besteditfadeout>`$3`$4`n`t`t`t`t`t`t<besteditfadein>$tmplng3</besteditfadein>"
                                    }
                                    else
                                    {
                                        $speedclip = $speedclip -replace '(?ms)(<clipitem ((?!>).)*?>)(.*)(<transitionitem>)', "`$1`n`t`t`t`t`t`t<besteditfadeout>$tmplng3</besteditfadeout>`$3`$4"
                                    }
                                    break
                                }

                                'end' {
                                    if ($tmpcenter -eq $true)
                                    {
                                        $speedclip = $speedclip -replace '(?ms)(<clipitem ((?!>).)*?>)(.*)(<clipitem ((?!>).)*?>)', "`$1`n`t`t`t`t`t`t<besteditfadeout>$tmplng3</besteditfadeout>`$3`$4`n`t`t`t`t`t`t<besteditfadein>$tmplng3</besteditfadein>"
                                    }
                                    else
                                    {
                                        $speedclip = $speedclip -replace '(?ms)(<clipitem ((?!>).)*?>)(.*)(<transitionitem>)', "`$1`n`t`t`t`t`t`t<besteditfadeout>$tmplng3</besteditfadeout>`$3`$4"
                                    }
                                    break
                                }

                                'start-black' {
                                    if ($tmpcenter -eq $true)
                                    {
                                        # this should never be true, but just in case...
                                        Write-Output "Warning: Start-Black conversion anomaly at '$tmplng'"
                                        $speedclip = $speedclip -replace '(?ms)(<clipitem ((?!>).)*?>)(.*)(<clipitem ((?!>).)*?>)', "`$1`n`t`t`t`t`t`t<besteditfadeout>$tmplng3</besteditfadeout>`$3`$4`n`t`t`t`t`t`t<besteditfadein>$tmplng3</besteditfadein>"
                                    }
                                    else
                                    {
                                        $speedclip = $speedclip -replace '(?ms)(</transitionitem>)(.*)(<clipitem ((?!>).)*?>)', "`$1`$2`$3`n`t`t`t`t`t`t<besteditfadein>$tmplng3</besteditfadein>"
                                    }
                                    break
                                }

                                'start' {
                                    if ($tmpcenter -eq $true)
                                    {
                                        $speedclip = $speedclip -replace '(?ms)(<clipitem ((?!>).)*?>)(.*)(<clipitem ((?!>).)*?>)', "`$1`n`t`t`t`t`t`t<besteditfadeout>$tmplng3</besteditfadeout>`$3`$4`n`t`t`t`t`t`t<besteditfadein>$tmplng3</besteditfadein>"
                                    }
                                    else
                                    {
                                        $speedclip = $speedclip -replace '(?ms)(</transitionitem>)(.*)(<clipitem ((?!>).)*?>)', "`$1`$2`$3`n`t`t`t`t`t`t<besteditfadein>$tmplng3</besteditfadein>"
                                    }
                                    break
                                }

                            }

                            # this serves two purposes, to mark effects we processed, and also setting it to an invalid effectid will make Premiere Pro import it as Constant Power rather than Constant Gain, and we want these overlay transitions as Constant Power
                            $speedclip = $speedclip -replace '(?ms)(<effectid>KGAudioTransCrossFade0dB</effectid>)', "<effectid>KGAudioTransCrossFadeXdB</effectid>"

                            #Write-Output "speedmatch regexF '$i', '$speedclip'"

                            # put our reprocessed clip back into the place of the original
                            #$wholefile = $wholefile -replace $speedclip_orig, $speedclip
                            $wholefile = $wholefile.Replace($speedclip_orig, $speedclip)
                        }
                    }

                }

                #Write-Output "wholefile: '$wholefile'"

                $outstream = New-Object IO.StreamWriter $newname

                # no longer necessary, we get this when we get the framerate now
                #$fileticksperframe = (24/$fileframerate)*$ppro24fps

                # recalculate cutpointshift to fix Adobe import rounding errors
                # subtract 0.001 to round the frames to the biggest possible frame size to compensate for the rounded off last 3 digits
                # since 23.976 is a bigger frame than 24 (less of them per second), Premiere Pro doesn't think we adjusted our cutpointticks properly if we only subtract that much to shift a frame over
                # we need to exaggerate the value slightly using this ratio so that we're a bit over 1 frame by the original values, so that when the original framerate clip is reimported it doesn't auto-center our transition when it should only be 1 frame from the end
                # using a larger shift like 0.002 doesn't seem to help the remaining transitions that are still centered, so those aren't a minor rounding error
                # $cutpointshift = $ppro24fps * (24/($fileframerate - 0.001))

                # now that we're using ticks rather than framerate it doesn't seem like this extra margin is necessary
                $cutpointshift = $pproconst

                $sequence_initialized = $false
                $in_filter = $false
                $filtertype = -1
                $filtercache = ""
                $filtervalue = ""
                [long]$tmp_ppro_in = -2
                [long]$tmp_ppro_in_frames = -2
                [long]$tmp_ppro_out = -2
                [long]$tmpduration = -1 # have to predeclare this as a long or the function return values cause problems

                [long]$tmp_frametotal = -1
                [long]$tmp_ticktotal = -1
                [long]$tmplng = 0
                [long]$tmplng2 = 0
                [decimal]$tmpdec = 0
                [decimal]$clipspeed = -1
                [long]$cliplength = -1

                # new caching configuration now that we have to adjust almost every clip and transition, and need to be able to look ahead at the ticks to do it
                $lookaheadon = $false
                $lookaheadcache = ""
                $shiftnextclip = $false
                $lastclipshifted = $false
                [long]$shiftcentercheck = -1

                # new vars for exponential fade
                # we're defining this as globals even though they should really be passed by ref since powershell byref is absolute garbage, convoluted and messy code
                [long]$global:expfadein = 0
                [long]$global:expfadeout = 0
                [long]$global:expcacheinedge = -1
                [long]$global:expcacheininner = -1
                [long]$global:expcacheoutedge = -1
                [long]$global:expcacheoutinner = -1

                # resplit lines since we may have changed some with previous processing
                $wholefile_lines = $wholefile -split "`n"
    
                #[System.IO.File]::ReadLines($srcpath + $origname1) | ForEach-Object {
                ForEach ($linedata_orig in $wholefile_lines) {

                    #$linedata_orig = $_

                    switch -regex ($linedata_orig)
                    {
                        # found a <name> tag
                        '(<name>)(.*)(</name>)' { 
                            if ($firstname -eq $true)
                            {
                                if ($cleanfilenames -eq $true)
                                {
                                    # this is the project title, not a potential file name, fix it to match the save file
                                    $linedata_orig = $linedata_orig -replace '(?<num1>.+<name>)(.*)(</name>)', "`${num1}$newtitle`$2"
                                }
                                
                                $firstname = $false
                            }
                            elseif ($in_filter -eq $true -and $filtertype -eq -1)
                            {
                                # filters have names, we can use this to identify and change the bad ones like Gain that don't reimport correctly
                                $tmpfilter = $linedata_orig -replace '(<name>)(.*)(</name>)', "`$2"
                                $tmpfilter = $tmpfilter.ToLower().Trim()

                                switch ($tmpfilter)
                                {
                                    'gain' {
                                        $filtertype = $filtergainconst
                                    }

                                    'audio levels' {
                                        $filtertype = $filterlevelconst
                                    }
                                }
                            }
                            elseif ($linedata_orig -match '(.*<name>)(.*)(1080p|720p)(.*)(</name>.*)')
                            {
                                if ($cleanfilenames -eq $true)
                                {
                                    # normalize name first
                                    $linedata_orig = ($Matches.2 + $Matches.3 + $Matches.4) -replace '[\[\]\(\)_]', ' ' # stage 1, replace crap with spaces
                                    $linedata_orig = $linedata_orig -replace '\s+', '.' # stage 2, replace multiple spaces in a row with one period
                                    $linedata_orig = $Matches.1 + $linedata_orig + $Matches.5 # reassemble tags

                                    if ($linedata_orig -match '(.+)(s[0-9]{2}e[0-9]{2})(.*)(1080p|720p)(.*)(.{4}</name>)') # try a secondary more detailed match for tv episodes
                                    {
                                        $linedata_orig = $Matches.1 + $Matches.2 + "." + $Matches.4 + $Matches.6
                                    }
                                    elseif ($linedata_orig -match '(.+)([0-9]{4})(.*)(1080p|720p)(.*)(.{4}</name>)') # try a secondary more detailed match for movies with year to clean up DTS, etc
                                    {
                                        $linedata_orig = $Matches.1 + $Matches.2 + "." + $Matches.4 + $Matches.6
                                    }
                                    else
                                    {
                                        $linedata_orig = $linedata_orig -replace '(.+)(1080p|720p)(.*)(.{4}</name>)', '$1$2$4'
                                    }
                                }
                            }
                            break
                        }

                        '(<logginginfo>)' { 
                            $isstrip++
                            break
                        }
                        '(</logginginfo>)' { 
                            $linedata_orig = '' # nuke this line so we don't insert the end tag
                            $isstrip--
                            break
                        }

                        '(<uuid>)' { 
                            $linedata_orig = '' # nuke this line
                            break
                        }

                        '(<pathurl>)(.*)(</pathurl>)' { 
                            if ($cleanfilenames -eq $true)
                            {
                                # with multiple source files we can't just use the cleaned name, have to recalculate it
                                if ($linedata_orig -match '(.*<pathurl>)(.*/)(((?!/).)*)(1080p|720p)(((?!/).)*)(</pathurl>.*)')
                                {
                                    # careful with the negated capture groups, they create a second match of the last character in their match
                                    #foreach ($tmppair in $Matches.GetEnumerator())
                                    #{
                                    #    Write-Output ("matchdata: '" + $tmppair.Name + "','" + $tmppair.Value + "'")
                                    #}

                                    # normalize name first
                                    $linedata_orig = ($Matches.3 + $Matches.5 + $Matches.6) -replace '[\[\]\(\)_]', ' ' # stage 1, replace crap with spaces
                                    $linedata_orig = $linedata_orig -replace '\s+', '.' # stage 2, replace multiple spaces in a row with one period
                                    $linedata_orig = $Matches.1 + $cleanpath + $linedata_orig + $Matches.8 # reassemble tags

                                    if ($linedata_orig -match '(.+)(s[0-9]{2}e[0-9]{2})(.*)(1080p|720p)(.*)(.{4}</pathurl>)') # try a secondary more detailed match for tv episodes
                                    {
                                        $linedata_orig = $Matches.1 + $Matches.2 + "." + $Matches.4 + $Matches.6
                                    }
                                    elseif ($linedata_orig -match '(.+)([0-9]{4})(.*)(1080p|720p)(.*)(.{4}</pathurl>)') # try a secondary more detailed match for movies with year to clean up DTS, etc
                                    {
                                        $linedata_orig = $Matches.1 + $Matches.2 + "." + $Matches.4 + $Matches.6
                                    }
                                    else
                                    {
                                        $linedata_orig = $linedata_orig -replace '(.+)(1080p|720p)(.*)(.{4}</pathurl>)', '$1$2$4'
                                    }
                                }
                                else
                                {
                                    # not end of the world but if we use the cleanname the auto-relink in Premiere won't work properly
                                    $linedata_orig = $linedata_orig -replace '(.+<pathurl>)(.*)(.{4}</pathurl>)', "`$1$cleanpath$cleanname`$3"
                                }
                            }
                            break
                        }

                        '(<sequence )' { 
                            # cull all the crap off the sequence tag
                            # the explodedtracks tag is now required for 2021 projects to load correctly
                            $linedata_orig = $linedata_orig -replace '(.*<sequence)(.*)( id="[^"]+")(.*)( explodedTracks="[^"]+")(.*)', "`$1`$3`$5>"

                            # reset for each new sequence
                            $sequence_initialized = $false
                            break
                        }

                        '(<timebase>)(.*)(</timebase>)' { 
                            # convert all timebase to the rounded framerate
                            # capture values shifted due to $num1, so $2 is actually $3 originally
                            $linedata_orig = $linedata_orig -replace '(?<num1><timebase>)(.*)(</timebase>)', "`${num1}$pprotimebase`$2"
                            break
                        }

                        '(<ntsc>)(.*)(</ntsc>)' { 
                            # convert all ntsc to false
                            $linedata_orig = $linedata_orig -replace '(<ntsc>)(.*)(</ntsc>)', "`$1$pprontsc`$3"
                            break
                        }

                        '(<duration>)(.*)(</duration>)' { 
                            $tmpduration = $linedata_orig -replace '(<duration>)(.*)(</duration>)', "`$2"
                            $tmpduration = ConvertFrameCount($tmpduration)
                            $linedata_orig = $linedata_orig -replace '(?<num1><duration>)(.*)(</duration>)', "`${num1}$tmpduration`$2"
                            break
                        }

                        '(<start>)(.*)(</start>)' { 
                            # turn the lookahead on when we hit any start value and turn it off when we get to the cutpointticks in a transitionitem, or pproticksout in a clipitem
                            # different transitions may or may not have pproticksin/out, seems like video transitions often don't have them
                            # ****
                            $lookaheadon = $true

                            # if this is -1 it indicates we had a prior transition
                            $tmpduration = ($linedata_orig -replace '(<start>)(.*)(</start>)', "`$2").Trim()

                            # if it's marked -1 it's always going to be that value, since that indicates it's connected to a transition
                            if ($tmpduration -ne "-1")
                            {
                                # need to save some values if we're in a transition so we can recalculate the frame count
                                if ($lookaheadon -eq $true -and $currentfixtype -eq $fixtransconst)
                                {
                                    $tmp_frametotal = [System.Convert]::ToInt64($tmpduration)
                                }

                                $tmpduration = ConvertFrameCount($tmpduration)

                                # if this is a transition we need to cache the start/end converted ticks
                                if ($currentfixtype -eq $fixtransconst)
                                {
                                    $tmp_ppro_in = CalculatePproTicks($tmpduration) # cache for later, saves work
                                    $tmp_ppro_in_frames = $tmpduration
                                }

                                $linedata_orig = $linedata_orig -replace '(?<num1><start>)(.*)(</start>)', "`${num1}$tmpduration`$2"
                            }
                            elseif ($currentfixtype -eq $fixtransconst)
                            {
                                # need to save some values if we're in a transition so we can recalculate the frame count
                                if ($lookaheadon -eq $true)
                                {
                                    $tmp_frametotal = -1
                                }
                            }
                            break
                        }

                        '(<end>)(.*)(</end>)' { 
                            # if this is -1 it indicates we have an upcoming transition
                            $tmpduration = ($linedata_orig -replace '(<end>)(.*)(</end>)', "`$2").Trim()

                            # if it's marked -1 it's always going to be that value, since that indicates it's connected to a transition
                            if ($tmpduration -ne "-1")
                            {
                                if ($lookaheadon -eq $true -and $currentfixtype -eq $fixtransconst)
                                {
                                    $tmp_frametotal = [System.Convert]::ToInt64($tmpduration) - $tmp_frametotal
                                }

                                $tmpduration = ConvertFrameCount($tmpduration)

                                # if this is a transition we need to cache the start/end converted ticks
                                if ($currentfixtype -eq $fixtransconst)
                                {
                                    $tmp_ppro_out = CalculatePproTicks($tmpduration) # cache for later, saves work
                                }

                                $linedata_orig = $linedata_orig -replace '(?<num1><end>)(.*)(</end>)', "`${num1}$tmpduration`$2"
                            }
                            elseif ($currentfixtype -eq $fixtransconst)
                            {
                                if ($lookaheadon -eq $true)
                                {
                                    $tmp_frametotal = -1
                                }
                            }
                            break
                        }

                        '(<alignment>)(.*)(</alignment>)' { 

                            # the end transition type seems to be the only one that is broken
                            # we need to change it to a middle with transition with one frame overlap

                            $tmpalign = $linedata_orig -replace '(<alignment>)(.*)(</alignment>)', "`$2"
                            $tmpalign = $tmpalign.ToLower().Trim()

                            if ($tmpalign -eq "end")
                            {
                                # change the alignment type
                                # WARNING: be careful, if you change the alignment to CENTER without adjusting the associated cutpointticks to a non-end value Premiere Pro will just automatically stick the cut in the very center of your transition, nowhere near the end.
                                # that must be a default protection feature in case the cutpointticks value is slightly outside the expected on either end, so make sure you shift at least one frame from the end to make sure rounding errors don't screw up your transitions
                                
                                # Version 11 fix, commenting this out so sequence framerate changes don't break, but we're keeping the shift code intact, it seems fine.  
                                # May be slightly inaccurate but better than the completely inaccurate conversion Premiere does if we don't mark it END
                                $linedata_orig = $linedata_orig -replace '(?<num1><alignment>)(.*)(</alignment>)', "`${num1}center`$2"

                                $shiftnextclip = $true # need to set this so that when we turn off the caching we still fix the next one
                            }

                            # for transitions we're using the alignment tag to end the caching
                            if ($currentfixtype -eq $fixtransconst)
                            {
                                # before anything else we need to recalculate the transition frame count and make sure it doesn't have any errors
                                if ($tmp_frametotal -ne -1 -and $tmp_ticktotal -ne -1)
                                {
                                    $tmp_ticktotal = [math]::Round($tmp_ticktotal/$pproconst)

                                    if ($tmp_ticktotal -gt $tmp_frametotal)
                                    {
                                        # technically we're doing this twice rather than just using pproticksin for all start/end, but sometimes we don't have all the tags so this is an easy way to be comprehensive without getting too complicated
                                        $tmplng = ($lookaheadcache -replace '(?ms)(.*<pproTicksIn>)(.*)(</pproTicksIn>.*)', "`$2").Trim()
                                        $tmplng = [math]::Round($tmplng/$pproconst)
                                        $lookaheadcache = $lookaheadcache -replace '(?<num1><start>)(.*)(</start>)', "`${num1}$tmplng`$2"

                                        $tmplng2 = ($lookaheadcache -replace '(?ms)(.*<pproTicksOut>)(.*)(</pproTicksOut>.*)', "`$2").Trim()
                                        $tmplng2 = $tmplng + $tmp_ticktotal
                                        $lookaheadcache = $lookaheadcache -replace '(?<num1><end>)(.*)(</end>)', "`${num1}$tmplng2`$2"

                                        $tmplng = CalculatePproTicks($tmplng)
                                        $lookaheadcache = $lookaheadcache -replace '(?<num1><pproTicksIn>)(.*)(</pproTicksIn>)', "`${num1}$tmplng`$2"

                                        
                                    }
                                }

                                # important that this comes *after* the previous function which corrects the START values
                                if ($tmpalign -eq "center")
                                {
                                    # look for bugs caused by END transitions erroneously marked as CENTER by PP rounding errors, are the cutpointticks values correct?

                                    # note that we're checking $linedata_orig and not the cache because we just wiped the cache
                                    $tmplng = ($lookaheadcache -replace '(?ms)(.*<start>)(.*)(</start>.*)', "`$2").Trim()
                                    $tmplng2 = ($lookaheadcache -replace '(?ms)(.*<end>)(.*)(</end>.*)', "`$2").Trim()
                                    # now we have the length of the transition to compare with the cutpointticks and see if it's subtracting the entire length
                                    # if it is we have a rogue END transition masquerading as a CENTER transition
                                    # important to fix, if we don't then the split will be error-corrected to the center of the transition by Premiere Pro, resulting in a completely different effect
                                    $shiftcentercheck = ($tmplng2 - $tmplng)
                                }

                                if ($centertransitions -eq $true -and $tmpalign -ne "start-black" -and $tmpalign -ne "end-black")
                                {
                                    # just a small patch, going to overwrite some of the stuff we just calculated if the user wants it centered
                                    # we shouldn't need to change the start or end of the transition since we want it in the same spot
                                    # and the length of the underlying clips should already overlap the entire transition regardless of the alignment
                                    # so we should just be able to change the type and the position of the cutpointticks
                                    #
                                    # ignoring any -black transitions since we won't be able to center those
                                    #
                                    $linedata_orig = $linedata_orig -replace '(?<num1><alignment>)(.*)(</alignment>)', "`${num1}center`$2"
                                    $shiftnextclip = $false # don't need to shift END transitions if we're already centering them
                                }


                                # we're done, clear the values
                                $tmp_frametotal = -1
                                $tmp_ticktotal = -1
                                $currentfixtype = -1 
                                
                            }
                            break
                        }

                        '(<cutPointTicks>)(.*)(</cutPointTicks>)' { 
     
                            # cutpointticks doesn't have any associate frame field so we need to use the known framerate to work backward from the ticks to find the framecount then calculate the fixed ticks from that
                            $tmpduration = $linedata_orig -replace '(<cutPointTicks>)(.*)(</cutPointTicks>)', "`$2"
                            $tmpduration = CalculatePproTicks(CalcConvFramesFromTicks($tmpduration))

                            if ($shiftcentercheck -ne -1 -and $tmpduration -eq 0)
                            {
                                # this is a second edge case, START transitions erroneously marked CENTER, retag it

                                # prevent it from being retagged with centertransitions true since only the start-black and end-black types are preserved, everything else is centered
                                if ($centertransitions -ne $true)
                                {
                                    $lookaheadcache = $lookaheadcache -replace '(?<num1><alignment>)(.*)(</alignment>)', "`${num1}start`$2"
                                }

                                $shiftcentercheck = -1
                            }
                            elseif ($shiftcentercheck -ne -1)
                            {
                                $tmplng = ($tmpduration/$pproconst)

                                if ($tmplng -eq $shiftcentercheck)
                                {
                                    # this is a rogue END transition marked as a CENTER, mark for fixing
                                    $shiftnextclip = $true

                                    #$tmpchange1 = ($lookaheadcache -replace '(?ms)(.*<end>)(.*)(</end>.*)', "`$2").Trim()
                                    #Write-Output "Before convert '$tmpchange1'"


                                    #$tmpchange2 = ConvertFramesToSrtTime($tmpchange1)
                                    #Write-Output "fixing rogue END: '$tmpchange2' (End: $tmpchange1)"

                                    # Version 11 fix: also change the alignment type to END since it isn't really a center
                                    # unfortunately fixing this causes Premiere to detach the transition from the left clip
                                    # I think the best option will be to expand the framerates used by the script so the sequence rate doesn't need to be changed
                                    # $lookaheadcache = $lookaheadcache -replace '(?<num1><alignment>)(.*)(</alignment>)', "`${num1}end`$2"
                                }

                                $shiftcentercheck = -1
                            }
                            elseif ($lookaheadcache -ne "")
                            {
                                # if it's not a center transition and we're caching, let's take a quick peek at the alignment type - if it's a one-sided transition 
                                # we need to make sure the cutpointticks lines up with the length of the transition or Premiere will just discard it.
                                $tmpstr = ($lookaheadcache -replace '(?ms)(.*<alignment>)(.*)(</alignment>.*)', "`$2").ToLower().Trim()

                                if ($tmpstr -eq "end-black")
                                {
                                    # video transitions don't have pproticksin/out, so for those we need to use the start/end values instead
                                    if ($lookaheadcache -match '(?ms)(.*<pproTicksIn>)(.*)(</pproTicksIn>.*)')
                                    {
                                        # grab our already corrected tick amounts
                                        $tmplng = ($lookaheadcache -replace '(?ms)(.*<pproTicksIn>)(.*)(</pproTicksIn>.*)', "`$2").Trim()
                                        $tmplng2 = ($lookaheadcache -replace '(?ms)(.*<pproTicksOut>)(.*)(</pproTicksOut>.*)', "`$2").Trim()
                                    }
                                    else
                                    {
                                        # grab our already corrected frame amounts and multiply
                                        $tmplng = ($lookaheadcache -replace '(?ms)(.*<start>)(.*)(</start>.*)', "`$2").Trim()
                                        $tmplng2 = ($lookaheadcache -replace '(?ms)(.*<end>)(.*)(</end>.*)', "`$2").Trim()

                                        $tmplng *= $pproconst
                                        $tmplng2 *= $pproconst
                                    }

                                    $tmpduration = $tmplng2 - $tmplng
                                }
                                elseif ($tmpstr -eq "start-black")
                                {
                                    $tmpduration = 0
                                }
                            }

                            if ($shiftnextclip -eq $true)
                            {
                                # this is an END we've converted to a CENTER
                                # shift the transition one frame left 
                                $tmpduration -= $cutpointshift
                                $shiftnextclip = $false

                                # this is to make sure the audio lines up to the video by adjusting it the opposite direction one frame
                                $lastclipshifted = $true 
                            }

                            if ($centertransitions -eq $true -and $tmpalign -ne "start-black" -and $tmpalign -ne "end-black" -and $lookaheadcache -ne "")
                            {
                                # final operation for our small patch, center the ticks if they aren't already
                                $tmpcenterlng = ($lookaheadcache -replace '(?ms)(.*<start>)(.*)(</start>.*)', "`$2").Trim()
                                $tmpcenterlng2 = ($lookaheadcache -replace '(?ms)(.*<end>)(.*)(</end>.*)', "`$2").Trim()

                                # e.g (10 - 5) / 2 = 2.5 frames rounded
                                $tmpcenterduration = [math]::Round(($tmpcenterlng2 - $tmpcenterlng) / 2.0)
                                # now convert back to ticks
                                $tmpduration = $tmpcenterduration * $pproconst
                            }

                            $linedata_orig = $linedata_orig -replace '(?<num1><cutPointTicks>)(.*)(</cutPointTicks>)', "`${num1}$tmpduration`$2"

                            # prepend any cached stuff and turn off caching
                            $linedata_orig = $lookaheadcache + $linedata_orig
                            $lookaheadcache = ""
                            $lookaheadon = $false
                            break
                        }

                        '(<in>)(.*)(</in>)' { 

                            $tmpduration = $linedata_orig -replace '(<in>)(.*)(</in>)', "`$2"

                            if ($clipspeed -ne -1)
                            {
                                # this fixes the in/out durations, which are drastically wrong when a speed modifier is added to the clip
                                # not sure why they use completely wrong integers rather than using decimals like rational programmers
                                $cliplength = $tmpduration
                                $tmpduration = $tmpduration * $clipspeed
                            }

                            # need to save some values if we're in a transition so we can recalculate the frame count
                            if ($lookaheadon -eq $true -and $currentfixtype -eq $fixclipconst)
                            {
                                $tmp_frametotal = [System.Convert]::ToInt64($tmpduration)
                            }

                            $tmpduration = ConvertFrameCount($tmpduration)

                            if ($lastclipshifted -eq $true) # fudging this was incorrect. since the clips are always extended to the full length of the entire transition regardless of where the cutpoint is, we don't want to change the IN values
                            {
                                #$tmpduration++
                            }

                            # if this is a video clip we need to cache the in/out converted ticks
                            if ($currentfixtype -eq $fixclipconst)
                            {
                                $tmp_ppro_in = CalculatePproTicks($tmpduration) # cache for later, saves work
                                $tmp_ppro_in_frames = $tmpduration
                            }

                            if ($convert_0db_to_fade -eq $true -and $currentfixtype -eq $fixclipconst -and $expfadein -ne 0)
                            {
                                # clip that has fade conversion turned on, capture the value
                                if ($expfadein -gt 0)
                                {
                                    $global:expcacheinedge = $tmpduration
                                    $global:expcacheininner = $tmpduration + $expfadein
                                }
                                elseif ($expfadein -lt 0)
                                {
                                    # reflip the sign we were using as a tmpcenter indicator
                                    # this is half the length of a normal transition, so we need to adjust forward accordingly
                                    $global:expfadein = -$expfadein

                                    $global:expcacheinedge = $tmpduration + $expfadein
                                    $global:expcacheininner = $tmpduration + $expfadein + $expfadein
                                }
                            }

                            $linedata_orig = $linedata_orig -replace '(?<num1><in>)(.*)(</in>)', "`${num1}$tmpduration`$2"
                            break
                        }

                        '(<out>)(.*)(</out>)' { 
                            
                            $tmpduration = $linedata_orig -replace '(<out>)(.*)(</out>)', "`$2"

                            if ($clipspeed -ne -1)
                            {
                                # going to subtract these invalid in/out values to find our frames then add that to the first one which we convert properly
                                # this avoids any rounding issues from the conversion in the easiest way. Fractions are very important here, easy to lose multiple frames i.e. 24 rounded down to 2 and then back to 20. 
                                $cliplength = ConvertFrameCount($tmpduration - $cliplength)

                                $tmpduration = ($lookaheadcache -replace '(?ms)(.*<in>)(.*)(</in>.*)', "`$2").Trim()
                                $tmpduration += $cliplength
                                $cliplength = -1
                            }
                            
                            if ($lookaheadon -eq $true -and $currentfixtype -eq $fixclipconst)
                            {
                                $tmp_frametotal = [System.Convert]::ToInt64($tmpduration) - $tmp_frametotal
                            }

                            if ($clipspeed -eq -1)
                            {
                                $tmpduration = ConvertFrameCount($tmpduration)
                            }

                            # if this is a video clip we need to cache the in/out converted ticks
                            if ($currentfixtype -eq $fixclipconst)
                            {
                                $tmp_ppro_out = CalculatePproTicks($tmpduration) # cache for later, saves work
                            }

                            if ($convert_0db_to_fade -eq $true -and $currentfixtype -eq $fixclipconst -and $expfadeout -ne 0)
                            {
                                # clip that has fade conversion turned on, capture the value
                                

                                # clip that has fade conversion turned on, capture the value
                                if ($expfadeout -gt 0)
                                {
                                    $global:expcacheoutedge = $tmpduration
                                    $global:expcacheoutinner = $tmpduration - $expfadeout
                                }
                                elseif ($expfadeout -lt 0)
                                {
                                    # reflip the sign we were using as a tmpcenter indicator
                                    # this is half the length of a normal transition, so we need to adjust forward accordingly
                                    $global:expfadeout = -$expfadeout

                                    $global:expcacheoutedge = $tmpduration - $expfadeout
                                    $global:expcacheoutinner = $tmpduration - $expfadeout - $expfadeout
                                }
                            }

                            $linedata_orig = $linedata_orig -replace '(?<num1><out>)(.*)(</out>)', "`${num1}$tmpduration`$2"
                            break
                        }

                        '(<pproTicksIn>)(.*)(</pproTicksIn>)' { 

                            # need to save some values if we're in a transition so we can recalculate the frame count
                            if ($lookaheadon -eq $true)
                            {
                                $tmpduration = ($linedata_orig -replace '(<pproTicksIn>)(.*)(</pproTicksIn>)', "`$2").Trim()
                                $tmp_ticktotal = [System.Convert]::ToInt64($tmpduration)
                                
                            }
                            
                            if ($tmp_ppro_in -ne -2)
                            {
                                if ($currentfixtype -eq $fixclipconst)
                                {
                                    # We need to make sure the in/pproticksin values only get rounded up, otherwise you end up with rogue flashes of frames on the timeline
                                    # I suspect this is due to Premiere Pro showing a frame if any percentage of it is present, 
                                    # so 34,313.3338451045 rounded down to 34,313 is adding 30% of the previous frame to the start of the clip and PP thinks 30% is enough to show that frame on the timeline for the first frame of the clip
                                    $tmplng = CalcConvFramesFromTicksUP($tmpduration)
                                    
                                    if ($tmplng -ne $tmp_ppro_in_frames -and $clipspeed -eq -1)
                                    {
                                        # must have been rounded down from the pproticksin value, fix it
                                        $lookaheadcache = $lookaheadcache -replace '(?<num1><in>)(.*)(</in>)', "`${num1}$tmplng`$2"
                                        $tmp_ppro_in = CalculatePproTicks($tmplng)
                                    }
                                }

                                $linedata_orig = $linedata_orig -replace '(?<num1><pproTicksIn>)(.*)(</pproTicksIn>)', "`${num1}$tmp_ppro_in`$2"
                                $tmp_ppro_in = -2
                                $tmp_ppro_in_frames = -2
                            }
                            break
                        }

                        '(<pproTicksOut>)(.*)(</pproTicksOut>)' { 
                            
                            [long]$tmpdiff = 0

                            # need to save some values if we're in a transition so we can recalculate the frame count
                            if ($lookaheadon -eq $true)
                            {
                                $tmpduration = ($linedata_orig -replace '(<pproTicksOut>)(.*)(</pproTicksOut>)', "`$2").Trim()
                                $tmp_ticktotal = [System.Convert]::ToInt64($tmpduration) - $tmp_ticktotal
                            
                                # recalculate out/end from pproticksout
                                $tmplng = [math]::Round($tmpduration/$fileticksperframe)
                                $tmplng = ConvertFrameCount($tmplng) # need to convert because the fields we'll be comparing with are already converted
                                
                                if ($currentfixtype -eq $fixclipconst)
                                {
                                    #clip
                                    if ($clipspeed -eq -1)
                                    {
                                        $tmplng2 = ($lookaheadcache -replace '(?ms)(.*<out>)(.*)(</out>.*)', "`$2").Trim()
                                
                                        if ($tmplng -ne $tmplng2)
                                        {
                                            # replace with the correct rounded ticks value, then recalculate the pproticks
                                            $lookaheadcache = $lookaheadcache -replace '(?<num1><out>)(.*)(</out>)', "`${num1}$tmplng`$2"

                                            $tmp_ppro_out = CalculatePproTicks($tmplng) # cache for later, saves work

                                            # calculate the difference
                                            $tmpdiff = ($tmplng - $tmplng2)

                                            # grab the end value to see if we want to fix it as well
                                            # **** will this cause overlap if we don't fix adjacent starts?
                                            $tmplng = ($lookaheadcache -replace '(?ms)(.*<end>)(.*)(</end>.*)', "`$2").Trim()

                                            if ($tmplng -ne "-1")
                                            {
                                                # subtract difference from end and then reassign it back
                                                $tmplng -= $tmpdiff
                                                $lookaheadcache = $lookaheadcache -replace '(?<num1><end>)(.*)(</end>)', "`${num1}$tmplng`$2"
                                            }
                                            else
                                            {
                                                $tmpdiff = 0
                                            }
                                        }
                                    }
                                }
                                else
                                {
                                    #transition
                                    $tmplng2 = ($lookaheadcache -replace '(?ms)(.*<end>)(.*)(</end>.*)', "`$2").Trim()
                                
                                    if ($tmplng -ne $tmplng2)
                                    {
                                        # replace with the correct rounded ticks value, then recalculate the pproticks
                                        $lookaheadcache = $lookaheadcache -replace '(?<num1><end>)(.*)(</end>)', "`${num1}$tmplng`$2"

                                        $tmp_ppro_out = CalculatePproTicks($tmplng) # cache for later, saves work
                                    }
                                }

                                if ($currentfixtype -eq $fixclipconst)
                                {
                                    # before anything else we need to recalculate the transition frame count and make sure it doesn't have any errors
                                    if ($tmp_frametotal -ne -1 -and $tmp_ticktotal -ne -1 -and $clipspeed -eq -1)
                                    {
                                        $tmp_ticktotal = [math]::Round($tmp_ticktotal/$fileticksperframe)

                                        if ($tmpdiff -ne 0)
                                        {
                                            # see if we need to subtract any changes we made to END values earlier in this function so we don't accidentally make the frames bigger on both ends
                                            $tmp_frametotal -= $tmpdiff
                                        }

                                        if ($tmp_ticktotal -gt $tmp_frametotal)
                                        {
                                            # technically we're doing this twice rather than just using pproticksin for all in/out, but sometimes we don't have all the tags so this is an easy way to be comprehensive without getting too complicated
                                            $tmplng = ($lookaheadcache -replace '(?ms)(.*<pproTicksIn>)(.*)(</pproTicksIn>.*)', "`$2").Trim()
                                            $tmplng = [math]::Round($tmplng/$pproconst)

                                            if ($lastclipshifted -eq $true) # fudging this was incorrect. since the clips are always extended to the full length of the entire transition regardless of where the cutpoint is, we don't want to change the IN values
                                            {
                                                #$tmplng++
                                            }

                                            # make sure we don't give clips negative values accidentally
                                            if ($tmplng -lt 0)
                                            {
                                                $tmplng = 0
                                            }

                                            $lookaheadcache = $lookaheadcache -replace '(?<num1><in>)(.*)(</in>)', "`${num1}$tmplng`$2"

                                            # now recalc the pproticksin from the new IN value, using duration for placeholder so we don't mess up the value
                                            $tmpduration = CalculatePproTicks($tmplng)

                                            $lookaheadcache = $lookaheadcache -replace '(?<num1><pproTicksIn>)(.*)(</pproTicksIn>)', "`${num1}$tmpduration`$2"

                                            # note we're in pproticksout now, so it isn't in the cache yet
                                            #$tmplng2 = ($linedata_orig -replace '(<pproTicksOut>)(.*)(</pproTicksOut>)', "`$2").Trim()
                                            $tmplng2 = $tmplng + $tmp_ticktotal

                                            # resave this value so we don't use an earlier incorrect one
                                            $tmp_ppro_out = CalculatePproTicks($tmplng2) # cache for later, saves work

                                            $lookaheadcache = $lookaheadcache -replace '(?<num1><out>)(.*)(</out>)', "`${num1}$tmplng2`$2"

                                            # fix the start value to match if it isn't in a transition
                                            # **** will this cause overlap if we don't fix adjacent ends?
                                            $tmplng = ($lookaheadcache -replace '(?ms)(.*<start>)(.*)(</start>.*)', "`$2").Trim()

                                            # change ticktotal to the difference, should typically be 1
                                            $tmp_ticktotal -= $tmp_frametotal

                                            if ($tmplng -ne "-1")
                                            {
                                                $tmplng -= $tmp_ticktotal

                                                # make sure we don't give clips negative values accidentally
                                                if ($tmplng -lt 0)
                                                {
                                                    $tmplng = 0
                                                }

                                                $lookaheadcache = $lookaheadcache -replace '(?<num1><start>)(.*)(</start>)', "`${num1}$tmplng`$2"
                                            }
                                        }
                                    }

                                    if ($clipspeed -ne -1)
                                    {
                                        # now that we're done with all the normal calculations we need to put the in/out back to an invalid value if we're dealing with an altered clip speed
                                        $tmplng = ($lookaheadcache -replace '(?ms)(.*<in>)(.*)(</in>.*)', "`$2").Trim()
                                        $tmplng = ($tmplng / $clipspeed)
                                        $lookaheadcache = $lookaheadcache -replace '(?<num1><in>)(.*)(</in>)', "`${num1}$tmplng`$2"
                                        # **** this out value is wrong I think, want to round up to make sure it's enough full frames though
                                        $tmplng = ($lookaheadcache -replace '(?ms)(.*<out>)(.*)(</out>.*)', "`$2").Trim()
                                        $tmplng = ($tmplng / $clipspeed)
                                        $lookaheadcache = $lookaheadcache -replace '(?<num1><out>)(.*)(</out>)', "`${num1}$tmplng`$2"
                                    }

                                    # we're done, clear the values
                                    $tmp_frametotal = -1
                                    $tmp_ticktotal = -1
                                    $currentfixtype = -1

                                    # prepend any cached stuff and turn off caching
                                    $linedata_orig = $lookaheadcache + $linedata_orig
                                    $lookaheadcache = ""
                                    $lookaheadon = $false

                                    $lastclipshifted = $false
                                }

                                # create the updated ticks version if we have it
                                if ($tmp_ppro_out -ne -2)
                                {
                                    $linedata_orig = $linedata_orig -replace '(?<num1><pproTicksOut>)(.*)(</pproTicksOut>)', "`${num1}$tmp_ppro_out`$2"
                                }
                            }
                            else
                            {
                                if ($tmp_ppro_out -ne -2)
                                {
                                    $linedata_orig = $linedata_orig -replace '(?<num1><pproTicksOut>)(.*)(</pproTicksOut>)', "`${num1}$tmp_ppro_out`$2"
                                }
                            }

                            $tmp_ppro_out = -2

                            # done with the clipspeed after altering in/out, clear it out so we don't accidentally impact any filters or anything that might share these tags
                            $clipspeed = -1
                            break
                        }

                        '(<clipitem )' { 
                            # reset the values for a new clip
                            $tmp_ppro_in = -2
                            $tmp_ppro_out = -2
                            $currentfixtype = $fixclipconst
                            $tmp_frametotal = -1
                            $tmp_ticktotal = -1  
                            $lookaheadon = $false
                            $lookaheadcache = ""
                            $shiftnextclip = $false
                            $shiftcentercheck = -1
                            $clipspeed = -1

                            $global:expfadein = 0
                            $global:expfadeout = 0
                            $global:expcacheinedge = -1
                            $global:expcacheininner = -1
                            $global:expcacheoutedge = -1
                            $global:expcacheoutinner = -1

                            if ($simulate_51_audio -eq $true)
                            {
                                $link_count = 0
                                $create_linkcache = $false
                                $link_template = ""
                                $link_latest = ""

                                # if it's audio, strip the type off the clipitem
                                # <clipitem id="clipitem-288" premiereChannelType="mono">
                                # <clipitem id="clipitem-213" premiereChannelType="stereo">
                                # <clipitem id="clipitem-87" premiereChannelType="5.1">
                                #
                                #$linedata_orig = $linedata_orig -replace '(.*)( premiereChannelType=".*")(.*)', "`$1`$3"

                                # actually we'll just correct it instead
                                $linedata_orig = $linedata_orig -replace '(.*)(?<num2> premiereChannelType=")(.*?)(".*)', "`$1`${num2}5.1`$3"

                                # grab the clip value for our hashtable
                                $latest_clip = $linedata_orig -replace '(.*id=")(.*?)(".*)', "`$2"
                            }
                            break
                        }

                        '(</clipitem>)' { 
                            $lookaheadon = $false # it shouldn't be on at this point, but just to be sure
                            $lookaheadcache = ""
                            $currentfixtype = -1

                            if ($expfadein -ne 0 -or $expfadeout -ne 0)
                            {
                                # We have required fades that weren't applied, sneak them in before the end of the clip
                                $gainresult = 1

                                #$linedata_orig = ConstructExponentialFade($gainresult,[ref]$expfadein,[ref]$expfadeout,[ref]$expcacheinedge,[ref]$expcacheininner,[ref]$expcacheoutedge,[ref]$expcacheoutinner) + $linedata_orig
                                $tmpstr = ConstructExponentialFade($gainresult) # test without concat
                                $linedata_orig = $tmpstr + $linedata_orig
                            }
                            break
                        }

                        '(<transitionitem>)' { 
                            # reset the values for a new clip
                            $tmp_ppro_in = -2
                            $tmp_ppro_out = -2
                            $currentfixtype = $fixtransconst
                            $tmp_frametotal = -1
                            $tmp_ticktotal = -1                            
                            $lookaheadon = $false
                            $lookaheadcache = ""
                            $shiftnextclip = $false
                            $shiftcentercheck = -1
                            $clipspeed = -1
                            break
                        }

                        '(</transitionitem>)' { 
                            $lookaheadon = $false # it shouldn't be on at this point, but just to be sure
                            $lookaheadcache = ""
                            $currentfixtype = -1
                            break
                        }

                        '(<filter>)' { 
                            # reset the values for a new clip
                            if ($in_filter -eq $true)
                            {
                                $filtercache = ""
                            }

                            $in_filter = $true
                            $filtertype = -1
                            $filtervalue = ""
                            break
                        }

                        '(</filter>)' { 
                            # save out cached filter data if we have any that qualifies
                            if ($in_filter -eq $true)
                            {
                                $linedata_orig = $filtercache + $linedata_orig
                                $filtercache = ""
                            }

                            switch ($filtertype)
                            {
                                $filtergainconst {
                                    # replacing the gain filter with the level filter
                                    if ($filtervalue -ne "")
                                    {
                                        # convert from decibels to decimal
                                        $gainresult = DecibelsToDecimal($filtervalue)
                                        $linedata_orig = ConstructExponentialFade($gainresult)
                                    }
                                    else
                                    {
                                        # error, missing filter value
                                        Write-Output "Error, missing gain filter value for conversion"
                                        # nuke filter tag rather than transmitting it with an error
                                        $linedata_orig = ""
                                    }
                                }

                                $filterlevelconst {

                                    if ($expfadein -eq 0 -and $expfadeout -eq 0)
                                    {
                                        # no fades to simulate, no need to alter anything, just let the filtercache we added to linedata be output
                                    }
                                    else
                                    {
                                        # need to add fades, we can use this level effect to attach the keyframes rather than adding our placeholder effect

                                        # cache old filterleveltext for a second, we want to assign the new one so the function accesses it
                                        $tmpstr = $filterleveltext

                                        # this whole filter is in linedata now, grab it
                                        $filterleveltext = $linedata_orig

                                        # extract the level value then insert the expected variables
                                        $tmpdec = $filterleveltext -replace '(?ms)(.*)(<name>Level</name>((?!<keyframe>).)*?)(<value>)(((?!<value>).)*?)(</value>)(.*)', "`$4"
                                        $filterleveltext = $filterleveltext -replace '(?ms)(<name>Level</name>((?!<keyframe>).)*?)(<value>)(((?!<value>).)*?)(</value>)', "`$1`$3FILTERVALUE`$6`n`t`t`t`t`t`t`t`t`tFILTERKEYFRAMES"
                                        
                                        $linedata_orig = ConstructExponentialFade ($tmpdec)

                                        # put the old filterleveltext back, we're done
                                        $filterleveltext = $tmpstr
                                    }
                                }
                            }

                            $in_filter = $false
                            $filtertype = -1
                            $filtervalue = ""
                            break
                        }

                        '(<value>)(.*)(</value>)' { 
                            
                            if ($in_filter -eq $true -and $filtertype -eq $filtergainconst)
                            {
                                # to correct the gain filter the only thing we need is the value to recreate a proper level filter instead
                                $filtervalue = $linedata_orig -replace '(<value>)(.*)(</value>)', "`$2"
                            }
                            break
                        }

                        '(<channelcount>)(.*)(</channelcount>)' { 
                            if ($simulate_51_audio -eq $true)
                            {
                                # convert all audio channel counts to 6 for dolby 5.1
                                $linedata_orig = $linedata_orig -replace '(?<num1><channelcount>)(.*)(</channelcount>)', "`${num1}6`$2"
                            }
                            break
                        }

                        '(<numOutputChannels>)(.*)(</numOutputChannels>)' { 
                            if ($simulate_51_audio -eq $true)
                            {
                                # convert all audio channel counts to 6 for dolby 5.1
                                $linedata_orig = $linedata_orig -replace '(?<num1><numOutputChannels>)(.*)(</numOutputChannels>)', "`${num1}6`$2"
                            }
                            break
                        }

                        '(<outputs>)' { 
                            if ($simulate_51_audio -eq $true)
                            {
                                # easiest just to delete the outputs and add our own default
                                $isstrip++
                                $print_audio_outputs = $true
                            }
                            break
                        }
                        '(</outputs>)' { 
                            if ($simulate_51_audio -eq $true)
                            {
                                $linedata_orig = '' # nuke this line so we don't insert the end tag
                                $isstrip--
                            }
                            break
                        }

                        '(<track )(.*)(currentExplodedTrackIndex)(.*)(totalExplodedTrackCount)' { 
                            if ($simulate_51_audio -eq $true)
                            {
                                $link_ht = @{}
                                $tmpht = ""

                                # audio track is going to look something like this originally:
                                # <track hardpan="TRUE" TL.SQTrackAudioKeyframeStyle="0" TL.SQTrackShy="0" TL.SQTrackExpandedHeight="25" TL.SQTrackExpanded="0" MZ.TrackTargeted="1" currentExplodedTrackIndex="0" totalExplodedTrackCount="6" premiereTrackType="5.1">
                                # stereo:
                                # <track TL.SQTrackAudioKeyframeStyle="0" TL.SQTrackShy="0" TL.SQTrackExpandedHeight="25" TL.SQTrackExpanded="0" MZ.TrackTargeted="1" PannerCurrentValue="0.5" PannerIsInverted="true" PannerStartKeyframe="-91445760000000000,0.5,0,0,0,0,0,0" PannerName="Balance" currentExplodedTrackIndex="0" totalExplodedTrackCount="2" premiereTrackType="Stereo">
                                # mono:
                                # <track TL.SQTrackAudioKeyframeStyle="0" TL.SQTrackShy="0" TL.SQTrackExpandedHeight="25" TL.SQTrackExpanded="0" MZ.TrackTargeted="1" PannerCurrentValue="0.5" PannerIsInverted="true" PannerStartKeyframe="-91445760000000000,0.5,0,0,0,0,0,0" PannerName="Balance" currentExplodedTrackIndex="0" totalExplodedTrackCount="1" premiereTrackType="Stereo">


                                if ($split_audio_channels -eq $true)
                                {
                                    # we need to get rid of the premiereTrackType tag to fix a bug where AME freaks out and fails to render the fixed video because it thinks a track is missing
                                    # the only downside is that it will show each channel on a track rather than hiding all the channels on one track, but it's worth it to avoid render failures
                                    $linedata_orig = $linedata_orig -replace '(.*)( premiereTrackType=".*")(.*)', "`$1`$3"
                                }
                                else
                                {
                                    $linedata_orig = $linedata_orig -replace '(.*)( premiereTrackType=".*")(.*)', "`$1 premiereTrackType=`"5.1`"`$3"
                                }

                                # Now we want to check how many tracks we have, and which track we're on
                                

                                # use ? so it isn't greedy, important here
                                $tmplng = ($linedata_orig -replace '(?ms)(.*currentExplodedTrackIndex=")(.*?)(".*)', "`$2").Trim()
                                $tmplng2 = ($linedata_orig -replace '(?ms)(.*totalExplodedTrackCount=")(.*?)(".*)', "`$2").Trim()

                                $link_expected = $tmplng2

                                # if there are 6 tracks we still want to reprocess them because unlinking the video and audio also unlinks the audio channels, so we need to recreate the links between all the 5.1 channels for easier editing later
                                # if there are more or less than 6 we need to nuke everything except track index 0 and then adjust copies of it to create 6 tracks to simulate a 5.1 file
                                if ($tmplng -eq 0)
                                {
                                    # start capturing the first channel track data so we can use it to create the other 5 simulated channels
                                    $caching_dolby_track = $true
                                    $physical_track_count++ # count it so we can renumber and link the dolby channels properly
                                }
                                else
                                {
                                    $isstrip++
                                    $nuking_track = $true
                                }
                            }
                            break
                        }

                        '(</track>)' { 
                            if ($simulate_51_audio -eq $true)
                            {
                                if ($nuking_track -eq $true)
                                {
                                    $linedata_orig = '' # nuke this line so we don't insert the end tag
                                    $isstrip--
                                    $nuking_track = $false
                                }
                                elseif($caching_dolby_track -eq $true)
                                {
                                    # if this was currentExplodedTrackIndex then we need to duplicate it to create the desired 6 tracks
                                    $firstdolbytrack += $linedata_orig

                                    $linedata_orig = '' # nuke this line so we don't double insert the end tag

                                    # mark the exploded track count as 6 for 5.1 audio
                                    $firstdolbytrack = $firstdolbytrack -replace '(<track )(.*)(?<num3>totalExplodedTrackCount=")(.*?)(")', "`$1`$2`${num3}6`$4"

                                    # get rid of any video links, not really necessary to have the video linked to the audio, causes additional complications and possible bugs
                                    # for some reason this version is greedy even though the wildcards shouldn't be:
                                    # $firstdolbytrack = $firstdolbytrack -replace '(?ms)(<link>.*?<mediatype>video</mediatype>.*?</link>)', ""
                                    # add extra negative lookahead so it doesn't greedily include extra links
                                    $firstdolbytrack = $firstdolbytrack -replace '(?ms)(<link>((?!<link>).)*?<mediatype>video</mediatype>.*?</link>)', ""

                                    # now we have one fully complete track cached, adjust the variables and output the next 5
                                    
                                    # Replace the trackindex with the appropriate number, 1-6
                                    # <trackindex>1</trackindex>
                                    # careful, the links have trackindexes too, we only want to change the one in the sourcetrack section
                                    
                                    $tmpstr = ""

                                    for ($i=1;$i -lt 7;$i++)
                                    {
                                        # zero based
                                        $j = $i-1

                                        $tmpstr = $firstdolbytrack -replace '(?ms)(?<num1>.*?<sourcetrack>.*?<trackindex>)(.*?)(</trackindex>.*?</sourcetrack>.*?)', "`${num1}$i`$2"
                                        # outputchannelindex is important, it determines which actual audio track is being pulled from the source file, so if this is 1 it's the Left audio track for instance
                                        $tmpstr = $tmpstr -replace '(?<num1>.*<outputchannelindex>)(.*?)(</outputchannelindex>.*)', "`${num1}$i`$2"

                                        
                                        $tmpstr = $tmpstr -replace '(<track )(.*)(?<num3>currentExplodedTrackIndex=")(.*?)(")', "`$1`$2`${num3}$j`$4"

                                        # expand each set of replacement values we saved earlier, and replace them in our newly generated tracks
                                        foreach ($tmppair in $link_ht.GetEnumerator())
                                        {
                                            $tmpkey = $tmppair.Name
                                            $tmpval = $tmppair.Value
                                            $tmparr = $tmpval.Split(",")
                                            $tmpval = $tmparr[$j]

                                            # only want to replace the clipitem IDs, if we do a broad replace it will hit the LINK values too which we don't want
                                            $tmpstr = $tmpstr -replace "(<clipitem )(.*)(id=`"$tmpkey`")", "`$1`$2`id=`"$tmpval`""
                                        }

                                        $linedata_orig += "`n$tmpstr"
                                    }

                                    

                                    $caching_dolby_track = $false
                                    $firstdolbytrack = ''
                                }
                            }
                            break
                        }

                        '(<link>)' { 
                            # get rid of all video links
                            if ($link_expected -lt 1)
                            {
                                # video track
                                $isstrip++
                            }

                            if ($simulate_51_audio -eq $true)
                            {
                                $link_latest = ""
                                $create_linkcache = $true
                            }
                            break
                        }

                        '(<linkclipref>)' { 
                            if ($simulate_51_audio -eq $true)
                            {
                                # need to cache these values so we can reconstruct tracks that we're nuking since it's easier to nuke and reconstruct all besides the zero track, rather than correcting the link values on each track plus potentially generating new ones
                                if ($tmpht -ne "")
                                {
                                    $tmpht += ","
                                }

                                # by this point we don't know if it's an audio or video clip, just add it and we can remove it at the end link tag if it isn't audio
                                $tmpht += $linedata_orig -replace '(.*<linkclipref>)(.*)(</linkclipref>.*)', "`$2"
                            
                            }
                            break
                        }

                        '(<trackindex>)' { 
                            if ($simulate_51_audio -eq $true)
                            {
                                if ($create_linkcache -eq $true)
                                {
                                    # this trackindex is inside a LINK, having trackindexes is necessary for the links to work between the channels
                                    # this isn't actually used for physically placing the channels on tracks, just for the LINK to work properly
                                    # if this is missing or misindexed then you won't get a nice link group for the channels, but everything else should work fine
                                    # ((zerobasedtracknum*6)+channelnum)
                                    $tmplng = (($physical_track_count*6)+($link_count+1))
                                    $linedata_orig = $linedata_orig -replace '(?<num1>.*<trackindex>)(.*?)(</trackindex>.*)', "`${num1}$tmplng`$2"
                                }
                            }
                            break
                        }

                        '(</link>)' { 
                            if ($link_expected -lt 1)
                            {
                                # video track
                                $linedata_orig = '' # nuke this line so we don't insert the end tag
                                $isstrip--
                            }

                            if ($simulate_51_audio -eq $true -and $link_expected -gt 0) # video tracks won't have any $link_expected value
                            {
                                if ($create_linkcache -eq $true)
                                {
                                    $link_latest += $linedata_orig
                                    $create_linkcache = $false
                                }

                                # don't want any sneaky linked videos matched/counted
                                if ($link_latest -match '<mediatype>audio</mediatype>')
                                {
                                    if ($link_template -eq "")
                                    {
                                        $link_template = $link_latest
                                    }

                                    # first we want to count the links, and start inserting manufactured links after the last one
                                    $link_count++

                                    #<link>
							        #<linkclipref>clipitem-141</linkclipref>
							        #<mediatype>audio</mediatype>
							        #<trackindex>3</trackindex>
							        #<clipindex>51</clipindex>
							        #<groupindex>1</groupindex>
						            #</link>

                                    #careful of video links interspersed:
						            #<link>
						            #<linkclipref>clipitem-210</linkclipref>
						            #<mediatype>video</mediatype>
						            #<trackindex>1</trackindex>
						            #<clipindex>1</clipindex>
						            #</link>

                                    if ($link_count -ge $link_expected)
                                    {
                                        for ($i=($link_count+1);$i -lt 7;$i++)
                                        {
                                            $tmptrack = (($physical_track_count*6)+$i) # ((zerobasedtracknum*6)+channelnum)
                                            #change linkclipref and trackindex
                                            # groupindex seems to always be one, and clipindex will always be the same as the template
                                            $tmpstr = $link_template -replace '(?<num1>.*<trackindex>)(.*?)(</trackindex>.*)', "`${num1}$tmptrack`$2"
                                            $tmpstr = $tmpstr -replace '(?<num1>.*<linkclipref>clipitem-)(.*?)(</linkclipref>.*)', "`${num1}$extra_clip_autonumber`$2"

                                            $tmpht += ",clipitem-$extra_clip_autonumber"
                                            $extra_clip_autonumber++

                                            $linedata_orig += "`n$tmpstr"
                                        }

                                        # need to save the hash values 
                                        $link_ht.add($latest_clip, $tmpht)
                                        $tmpht = ""

                                        $link_template = ""
                                    }
                                }
                                else
                                {
                                    # link item was added that isn't audio, we need to remove it so we don't get the wrong counts
                                    $tmplng = $tmpht.LastIndexOf(',')

                                    if ($tmplng -eq -1)
                                    {
                                        # probably only one value with no commas yet
                                        $tmpht = ""
                                    }
                                    else
                                    {
                                        $tmpht = $tmpht.Substring(0,$tmplng)
                                    }
                                }
                            }
                            break
                        }

                        '(<besteditspeed>)(.*)(</besteditspeed>)' { 
                            
                            $clipspeed = [System.Convert]::ToDecimal(($linedata_orig -replace '(<besteditspeed>)(.*)(</besteditspeed>)', "`$2").Trim())
                            # convert to percent, the original tag has it as a whole number
                            $clipspeed = $clipspeed / 100

                            $linedata_orig = '' # nuke this line so we don't insert our temp tag in the final project
                            break
                        }

                        '(<besteditfadein>)(.*)(</besteditfadein>)' { 
                            # grabbing the length of the portion of the transition we need to fade is only one half of the equation
                            # we need to offset the in/out values of the clipitem to get the desired keyframe <when> values - they're values relative to the clip, not absolute timeline values
                            $global:expfadein = CalcConvFramesFromTicks(($linedata_orig -replace '(<besteditfadein>)(.*)(</besteditfadein>)', "`$2").Trim())

                            if ($expfadein -eq 1)
                            {
                                # if it's a single frame increase it to two so the fade has enough keyframe space
                                $global:expfadein = $expfadein + 1
                            }
                            elseif ($expfadein -eq -1)
                            {
                                $global:expfadein = $expfadein - 1
                            }

                            $linedata_orig = '' # nuke this line so we don't insert our temp tag in the final project
                            break
                        }

                        '(<besteditfadeout>)(.*)(</besteditfadeout>)' { 
                            
                            $global:expfadeout = CalcConvFramesFromTicks(($linedata_orig -replace '(<besteditfadeout>)(.*)(</besteditfadeout>)', "`$2").Trim())

                            if ($expfadeout -eq 1)
                            {
                                # if it's a single frame increase it to two so the fade has enough keyframe space
                                $global:expfadeout = $expfadeout + 1
                            }
                            elseif ($expfadeout -eq -1)
                            {
                                $global:expfadeout = $expfadeout - 1
                            }

                            $linedata_orig = '' # nuke this line so we don't insert our temp tag in the final project
                            break
                        }

                    }

                    if ($isstrip -lt 1 -and $in_filter -eq $false -and $lookaheadon -eq $false)
                    {
                        if ($create_linkcache -eq $true)
                        {
                            $link_latest += "$linedata_orig`n"
                        }

                        if ($caching_dolby_track -eq $true)
                        {
                            # originally was going to write this out live while caching it, but it's actually cleaner to do certain fixes to this track with the others and then output it later
                            $firstdolbytrack += "$linedata_orig`n"
                        }
                        else
                        {
                            $outstream.WriteLine($linedata_orig)
                        }
                    }
                    elseif ($isstrip -lt 1 -and $in_filter -eq $true)
                    {
                        $filtercache += "$linedata_orig`n"
                    }
                    elseif ($isstrip -lt 1 -and $lookaheadon -eq $true -and $linedata_orig -ne "")
                    {
                        $lookaheadcache += "$linedata_orig`n"
                    }
                    elseif ($print_audio_outputs -eq $true)  # we're stripping the normal outputs tags at this point, but we want to print the substitute tags for dolby 5.1
                    {
                        $print_audio_outputs = $false
                        $outstream.WriteLine($dolby5output)
                    }

                }

                $outstream.Close()
            

                if ($cleanname -ne "")
                {
                    # current time with a bit of randomness to avoid getting the same number over and over since we process fast
                    $currenttime = [math]::Floor( (New-TimeSpan -Start (Get-Date "01/01/1970") -End (Get-Date)).TotalSeconds ) - (Get-Random) 

                    # fix negative numbers
                    if ($currenttime -lt 0)
                    {
                        $currenttime = -$currenttime
                    }

                    # zip the created file and delete it
                    Compress-Archive -Path $newname -DestinationPath ($destpath + $newtitle + "_" + "premiere" + $pversion + "_" + $currenttime + ".zip")
                    Remove-Item -LiteralPath ($newname) -Force -Recurse

                    $donotmove = $false
                }
            }
            elseif ($isproject -eq $true)
            {
                Write-Output "Project detected, please select only one sequence in the project pane before selecting xml export! Skipping file: '$origname1'"
            }
            else
            {
                #Write-Output "Companion file not found: '$origname1NOext$companionext'"
                Write-Output "Framerate could not be detected: '$origname1'"
            }
            
        }
        else
        {
            Write-Output "File '$origname1' not in normal movie format (Name.Year.Resolution) or tv format (Name.S##E##.Resolution), skipping!"
        }

        if ($donotmove -eq $false)
        {
            Get-Item -LiteralPath ( $srcpath + $origname1 ) | Move-Item -Force -Destination { $finpath + $origname1 }
            #Get-Item -LiteralPath ( $srcpath + $origname1NOext + $companionext ) | Move-Item -Force -Destination { $finpath + $origname1NOext + $companionext }
        }
    }

    # save out any changes to our version database
    if ($add_version -eq $true)
    {
        $voutstream = New-Object IO.StreamWriter $versiondb

        foreach ($tmppair in $versions_ht.GetEnumerator())
        {
            $voutstream.WriteLine(($tmppair.Name + ',' + $tmppair.Value))
        }

        $voutstream.Close()
    }

    # save out any changes to our source id database
    if ($cleanvidid -eq $true)
    {
        $voutstream = New-Object IO.StreamWriter $srcviddb

        foreach ($tmppair in $srcvid_ht.GetEnumerator())
        {
            $voutstream.WriteLine(($tmppair.Name + ',' + $tmppair.Value))
        }

        $voutstream.Close()
    }
}


